"""
Sudoers - files ``/etc/sudoers`` and ``/etc/sudoers.d/*``
=========================================================

Module for processing each of the ``/etc/sudoers`` and ``/etc/sudoers.d/*`` files.

.. note::
    These files is filtered to skip the sensitive information.

.. note::
    Please use the :py:class:`insigths.combiners.suoders.Sudoers` for global
    checking.

"""
from insights.core import Parser
from insights.core.exceptions import SkipComponent
from insights.core.filters import add_filter
from insights.core.plugins import parser
from insights.parsers import get_active_lines
from insights.specs import Specs

add_filter(Specs.sudoers, "#includedir")
"""The "#includedir" is the line that should be always collected. """


class SudoersBase(object):
    """
    Base class for parsing the files ``/etc/sudoers`` or ``/etc/sudoers.d/*``,
    it provides the following two helper functions ``get`` and ``last``.
    """
    def get(self, s, check=all):
        """
        Returns all lines that contain `s` anywhere and return the list of RAW
        line directly. `s` can be either a single string or a string list. For
        list, all keywords in the list must be found in each line.

        Parameters:
            s(str or list): one or more strings to search for
            check(func): built-in function ``all`` or ``any`` applied to each line

        Returns:
            (list): list of lines that contain the `s`.

        Raises:
            TypeError: When `s` is not a string or a list of strings, or `num`
                is not an integer.
        """

        def _valid_search(s, check=all):
            if isinstance(s, str):
                return lambda l: s in l
            elif (isinstance(s, (tuple, list)) and len(s) > 0 and
                  all(isinstance(w, str) for w in s)):
                return lambda l: check(w in l for w in s)
            raise TypeError('"s" must be given as a string or a list of strings')

        search_by_expression = _valid_search(s, check)
        ret = []
        for _l in self.lines:
            if search_by_expression(_l):
                ret.append(_l)
        return ret

    def last(self, s, check=all):
        """
        Returns the last line that contain `s` anywhere and return the RAW
        line directly. `s` can be either a single string or a string list. For
        list, all keywords in the list must be found in each line.

        Parameters:
            s(str or list): one or more strings to search for
            check(func): built-in function ``all`` or ``any`` applied to each line

        Returns:
            (str): The line that contains the `s`. None by default.

        Raises:
            TypeError: When `s` is not a string or a list of strings, or `num`
                is not an integer.
        """
        ret = self.get(s, check)
        return ret[-1] if ret else None


@parser(Specs.sudoers)
class EtcSudoers(Parser, SudoersBase):
    """
    Class to parse the files ``/etc/sudoers`` or ``/etc/sudoers.d/*``

    Typical content of the  ``/etc/sudoers`` and ``/etc/sudoers.d/*`` is::

        ## Allows people in group wheel to run all commands
        %wheel  ALL=(ALL)       ALL
        ## Read drop-in files from /etc/sudoers.d (the # here does not mean a comment)
        #includedir /etc/sudoers.d

    Attributes:
        lines(list): The list of RAW lines of the file.

    .. note::
        The super-class :class:`SudoersBase` providers two helper functions:
        :func:`SudoersBase.get()` and :func:`SudoersBase.last()`.


    Examples:
        >>> type(sudo)
        <class 'insights.parsers.sudoers.EtcSudoers'>
        >>> len(sudo.lines)
        2
        >>> sudo.get(['wheel', 'ALL=(ALL)', 'ALL'])
        ['%wheel  ALL=(ALL)       ALL']
        >>> sudo.last("#includedir")
        '#includedir /etc/sudoers.d'
    """
    def parse_content(self, content):
        self.lines = get_active_lines(content, comment_char="##")

        if not self.lines:
            raise SkipComponent



from insights.parsers.sysconfig import DirsrvSysconfig
from insights.tests import context_wrap

SYSCONFIG = """
# how many seconds to wait for the startpid file to show
# up before we assume there is a problem and fail to start
# if using systemd, omit the "; export VARNAME" at the end
#STARTPID_TIME=10 ; export STARTPID_TIME
# how many seconds to wait for the pid file to show
# up before we assume there is a problem and fail to start
# if using systemd, omit the "; export VARNAME" at the end
#PID_TIME=600 ; export PID_TIME
KRB5CCNAME=/tmp/krb5cc_995
KRB5_KTNAME=/etc/dirsrv/ds.keytab
"""


def test_dirsrv_sysconfig():
    syscfg = DirsrvSysconfig(context_wrap(SYSCONFIG))
    # Standard access through data
    assert 'PID_TIME' not in syscfg.data
    assert syscfg.data['KRB5CCNAME'] == '/tmp/krb5cc_995'
    assert syscfg.data['KRB5_KTNAME'] == '/etc/dirsrv/ds.keytab'
    # pseudo-dictionary accessor
    assert syscfg['KRB5CCNAME'] == '/tmp/krb5cc_995'
    assert syscfg['KRB5_KTNAME'] == '/etc/dirsrv/ds.keytab'
    # No unparsed lines
    assert syscfg.unparsed_lines == []



"""
multipath.conf file content
===========================

The base class is the MultipathConfParser class, which reads the multipath
daemon's ``/etc/multipath.conf`` configuration file.
This is in a pseudo-JSON format.

MultipathConf - file ``/etc/multipath.conf``
--------------------------------------------

MultipathConfInitramfs - command ``lsinitrd -f /etc/multipath.conf``
--------------------------------------------------------------------

"""
import string

from insights.contrib import pyparsing as p
from insights.core import ConfigParser, LegacyItemAccess, Parser
from insights.core.exceptions import SkipComponent
from insights.core.plugins import parser
from insights.parsr import (EOF, Forward, LeftCurly, Lift, LineEnd, Literal,
                            Many, Number, OneLineComment, PosMarker, QuotedString,
                            RightCurly, String, WS, WSChar, skip_none)
from insights.parsr.query import Entry
from insights.specs import Specs


class MultipathConfParser(Parser, LegacyItemAccess):
    """
    Shared parser for the file ``/etc/multipath.conf`` and output of
    ``lsinitrd -f /etc/multipath.conf`` applied to
    /boot/initramfs-<kernel-version>.img.

    Return a dict where the keys are the name of sections in multipath
    configuraion file.  If there are subsections, the value is a list of
    dictionaries with parameters as key and value.  Otherwise the value is
    just a single dictionary.

    Configuration File Example::

        defaults {
               path_selector           "round-robin 0"
               user_friendly_names      yes
        }

        multipaths {
               multipath {
                       alias                   yellow
                       path_grouping_policy    multibus
              }
               multipath {
                       wwid                    1DEC_____321816758474
                       alias                   red
              }
        }

        devices {
               device {
                       path_selector           "round-robin 0"
                       no_path_retry            queue
              }
               device {
                       vendor                  1DEC_____321816758474
                       path_grouping_policy    red
              }
        }

        blacklist {
              wwid 26353900f02796769
              devnode "^hd[a-z]"
        }

    Parse Result::

        data = {
          "blacklist": {
            "devnode": "^hd[a-z]",
            "wwid": "26353900f02796769"
          },
          "devices": [
            {
              "path_selector": "round-robin 0",
              "no_path_retry": "queue"
            },
            {
              "path_grouping_policy": "red",
              "vendor": "1DEC_____321816758474"
            }
          ],
          "defaults": {
            "path_selector": "round-robin 0",
            "user_friendly_names": "yes"
          },
          "multipaths": [
            {
              "alias": "yellow",
              "path_grouping_policy": "multibus"
            },
            {
              "alias": "red",
              "wwid": "1DEC_____321816758474"
            }
          ]
        }
    """

    @classmethod
    def _create_parser(cls):
        """
        Need to check the specific symbol "/" in attr_value part as well.
        I checked some multipath configuraion files from the sosreport and got
        although there are some more specific symbols like "-%", it is enclosed
        in double quotes and will be accepted. Furthermore, I also checked the
        source code of "device-mapper-multipath" and got if the attr_value in
        "multipath.conf" include a "whitespace", it must be enclosed in double
        quotation marks. So, we could just add one more specific symbol "/" to
        check.
        ----------------------------------------------------------
        udev_dir                /dev
        getuid_callout          "/sbin/scsi_id -g -u -s /block/%n"
        ----------------------------------------------------------
        """
        section_name = p.Word(p.alphas + "_")
        attr_name = attr_value = p.Word(p.alphanums + "_/")
        LBRACE, RBRACE = map(p.Suppress, "{}")
        attr = p.Group(attr_name + (attr_value | p.quotedString.setParseAction(p.removeQuotes)))
        attr_list = p.Dict(p.ZeroOrMore(attr))
        simple_section = p.Group(section_name + LBRACE + attr_list + RBRACE)
        complex_section = p.Group(section_name + LBRACE + p.OneOrMore(simple_section) + RBRACE)
        simple_or_complex = p.Dict(simple_section | complex_section)
        my_conf = p.Group(p.ZeroOrMore(simple_or_complex))
        my_conf.ignore("#" + p.restOfLine)
        return my_conf

    def parse_content(self, content):
        if not content:
            raise SkipComponent("Empty content.")
        self.data = MultipathConfParser._create_parser().parseString("\n".join(content))[0].asDict()


@parser(Specs.multipath_conf)
class MultipathConf(MultipathConfParser):
    """
    Parser for the file ``/etc/multipath.conf``.

    Examples:
        >>> conf = shared[MultipathConf]
        >>> conf.data['blacklist']['devnode']  # Access via data property
        '^hd[a-z]'
        >>> conf['defaults']['user_friendly_names']  # Pseudo-dict access
        'yes'
        >>> len(conf['multipaths'])
        2
        >>> conf['multipaths'][0]['alias']
        'yellow'
    """
    pass


@parser(Specs.multipath_conf_initramfs)
class MultipathConfInitramfs(MultipathConfParser):
    """
    Parser for the output of ``lsinitrd -f /etc/multipath.conf`` applied to
    /boot/initramfs-<kernel-version>.img.

    Examples:
        >>> conf = shared[MultipathConfInitramfs]
        >>> conf.data['blacklist']['devnode']  # Access via data property
        '^hd[a-z]'
        >>> conf['defaults']['user_friendly_names']  # Pseudo-dict access
        'yes'
        >>> len(conf['multipaths'])
        2
        >>> conf['multipaths'][0]['alias']
        'yellow'
    """
    pass


def parse_doc(content, ctx):
    def to_entry(name, rest):
        if isinstance(rest, list):
            return Entry(name=name.value, children=rest, lineno=name.lineno, src=ctx)
        return Entry(name=name.value, attrs=[rest], lineno=name.lineno, src=ctx)

    Stmt = Forward()
    Num = Number & (WSChar | LineEnd)
    NULL = Literal("none", value=None)
    Comment = (WS >> OneLineComment("#").map(lambda x: None))
    BeginBlock = (WS >> LeftCurly << WS)
    EndBlock = (WS >> RightCurly << WS)
    Bare = String(set(string.printable) - (set(string.whitespace) | set("#{}'\"")))
    Name = WS >> PosMarker(String(string.ascii_letters + "_")) << WS
    Value = WS >> (Num | NULL | QuotedString | Bare) << WS
    Block = BeginBlock >> Many(Stmt).map(skip_none) << EndBlock
    Stanza = (Lift(to_entry) * Name * (Block | Value)) | Comment
    Stmt <= WS >> Stanza << WS
    Doc = Many(Stmt).map(skip_none)
    Top = Doc + EOF

    return Entry(children=Top(content)[0])


@parser(Specs.multipath_conf)
class MultipathConfTree(ConfigParser):
    """
    Exposes multipath configuration through the parsr query interface.

    See the :py:class:`insights.core.ConfigComponent` class for example usage.
    """
    def parse_doc(self, content):
        return parse_doc("\n".join(content), ctx=self)


def get_tree(root=None):
    """
    This is a helper function to get a multipath configuration component for
    your local machine or an archive. It's for use in interactive sessions.
    """
    from insights import run
    return run(MultipathConfTree, root=root).get(MultipathConfTree)


@parser(Specs.multipath_conf_initramfs)
class MultipathConfTreeInitramfs(ConfigParser):
    """
    Exposes the multipath configuration from initramfs image through the
    parsr query interface.

    See the :py:class:`insights.core.ConfigComponent` class for example usage.
    """
    def parse_doc(self, content):
        return parse_doc("\n".join(content), ctx=self)


def get_tree_from_initramfs(root=None):
    """
    This is a helper function to get a multipath configuration(from initramfs
    image) component for your local machine or an archive. It's for use in
    interactive sessions.
    """
    from insights import run
    return run(MultipathConfTreeInitramfs, root=root).get(MultipathConfTreeInitramfs)



"""
Sat6DBMigrateStatus - command ``foreman-rake db:migrate:status``
================================================================

This parser collects the output of the ``foreman-rake db:migrate:status``
command, which checks the status of all the migrations known to Foreman.
Each migration has a status, a date code, and a name.  These are stored
in a list of migrations, with 'up' migrations being listed in an ``up``
property and migrations with any other status being stored in a ``down``
property.

Sample input::

    database: foreman

     Status   Migration ID    Migration Name
    --------------------------------------------------
       up     20090714132448  Create hosts
       up     20090714132449  Add audits table
       up     20090715143858  Create architectures
       up     20090717025820  Create media
       up     20090718060746  Create domains
       up     20090718064254  Create subnets
       up     20090720134126  Create operatingsystems
       up     20090722140138  Create models

Examples:
    >>> status = shared[Sat6DBMigrateStatus]
    >>> status.database
    'foreman'
    >>> '20090714132448' in status.migrations
    True
    >>> '20090714140138' in status.migrations
    False
    >>> len(status.up)
    8
    >>> status.down
    []

"""
import re

from collections import namedtuple

from insights.core import CommandParser
from insights.core.exceptions import ParseException
from insights.core.plugins import parser
from insights.specs import Specs


Migration = namedtuple("Migration", ('status', 'id', 'name'))
"""namedtuple: Stores one migration record"""


@parser(Specs.foreman_rake_db_migrate_status)
class Sat6DBMigrateStatus(CommandParser):
    """
    Parse the ``foreman-rake db:migrate:status`` command.

    Attributes:
        database (str): The name of the database (usually 'foreman')
        migrations (dict): All the migrations, indexed by migration ID.
        up (list): Only the 'up' migrations, in order of appearance
        down (list): All migrations not listed as 'up', in order of appearance
    """

    def parse_content(self, content):
        self.database = ''
        self.migrations = {}
        self.up = []
        self.down = []
        mig_re = re.compile(r'\s+(?P<status>\w+)\s+(?P<id>\d+)\s+(?P<name>\w.*)$')
        for line in content:
            if line.startswith('database: '):
                self.database = line[len('database: '):]
                continue
            match = mig_re.search(line)
            if match:
                migration = Migration(*match.group('status', 'id', 'name'))
                self.migrations[match.group('id')] = migration
                if match.group('status') == 'up':
                    self.up.append(migration)
                else:
                    self.down.append(migration)

        # Did we actually get a database and some lines?  If not there must
        # have been a parse error.
        if not (self.database and self.migrations):
            raise ParseException("Could not find database name nor any migrations")



import json
import pytest

from mock.mock import Mock
from os import path

from insights.core.exceptions import SkipComponent
from insights.specs.datasources.leapp import leapp_report, LocalSpecs


with open(path.join(path.dirname(__file__), 'leapp-report.json'), 'r') as fp:
    INPUT_OK = fp.readlines()

with open(path.join(path.dirname(__file__), 'leapp-report.ret'), 'r') as fp:
    RESULT = fp.readlines()

INPUT_NG_1 = """
{ "entries":[
  {
    "a": "B",
  }
 ]
}
""".strip()

INPUT_NG_2 = ""


def test_leapp_report():
    report_file = Mock()
    report_file.content = INPUT_OK
    broker = {LocalSpecs.leapp_report_raw: report_file}

    result = leapp_report(broker)

    result_json = json.loads(''.join(result.content).strip())
    expected_json = json.loads(''.join(RESULT).strip())
    for ret in result_json:
        assert ret in expected_json

    report_file = Mock()
    report_file.content = INPUT_NG_1
    broker = {LocalSpecs.leapp_report_raw: report_file}
    with pytest.raises(SkipComponent):
        leapp_report(broker)

    report_file = Mock()
    report_file.content = INPUT_NG_2
    broker = {LocalSpecs.leapp_report_raw: report_file}
    with pytest.raises(SkipComponent):
        leapp_report(broker)



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import named_checkconf
from insights.parsers.named_checkconf import NamedCheckconf
from insights.tests import context_wrap


CONFIG_DNSSEC_ENABLED = """
options {
    bindkeys-file "/etc/named.iscdlv.key";
    directory "/var/named";
    dump-file "/var/named/data/cache_dump.db";
    listen-on port 53 {
        127.0.0.1/32;
    };
    listen-on-v6 port 53 {
        ::1/128;
    };
    managed-keys-directory "/var/named/dynamic";
    memstatistics-file "/var/named/data/named_mem_stats.txt";
    statistics-file "/var/named/data/named_stats.txt";
    dnssec-enable yes;
    dnssec-validation yes;
    recursion yes;
    allow-query {
        "localhost";
    };
};
"""

CONFIG_DNSSEC_DISABLED = """
options {
    bindkeys-file "/etc/named.iscdlv.key";
    directory "/var/named";
    dump-file "/var/named/data/cache_dump.db";
    listen-on port 53 {
        127.0.0.1/32;
    };
    listen-on-v6 port 53 {
        ::1/128;
    };
    managed-keys-directory "/var/named/dynamic";
    memstatistics-file "/var/named/data/named_mem_stats.txt";
    statistics-file "/var/named/data/named_stats.txt";
    dnssec-enable no;
    dnssec-validation yes;
    recursion yes;
    allow-query {
        "localhost";
    };
};
"""

CONFIG_DNSSEC_DEFAULT = """
options {
    bindkeys-file "/etc/named.iscdlv.key";
    directory "/var/named";
    dump-file "/var/named/data/cache_dump.db";
    listen-on port 53 {
        127.0.0.1/32;
    };
    listen-on-v6 port 53 {
        ::1/128;
    };
    managed-keys-directory "/var/named/dynamic";
    memstatistics-file "/var/named/data/named_mem_stats.txt";
    statistics-file "/var/named/data/named_stats.txt";
    dnssec-validation yes;
    recursion yes;
    allow-query {
        "localhost";
    };
};
"""

CONFIG_DISABLED_SECTIONS = """
logging {
    channel "default_debug" {
        file "data/named.run";
        severity dynamic;
    };
};
options {
    directory "/var/named";
    dump-file "/var/named/data/cache_dump.db";
    listen-on port 53 {
        127.0.0.1/32;
    };
    listen-on-v6 port 53 {
        ::1/128;
    };
    managed-keys-directory "/var/named/dynamic";
    memstatistics-file "/var/named/data/named_mem_stats.txt";
    pid-file "/run/named/named.pid";
    recursing-file "/var/named/data/named.recursing";
    secroots-file "/var/named/data/named.secroots";
    session-keyfile "/run/named/session.key";
    statistics-file "/var/named/data/named_stats.txt";
    disable-algorithms "." {
        "RSAMD5";
        "DSA";
    };
    disable-ds-digests "." {
        "GOST";
    };
    dnssec-enable yes;
    dnssec-validation yes;
    recursion yes;
    allow-query {
        "localhost";
    };
};
managed-keys {
    "." initial-key 257 3 8 "AwEAAagAIKlVZrpC6Ia7gEzahOR+9W29euxhJhVVLOyQbSEW0O8gcCjF
                FVQUTf6v58fLjwBd0YI0EzrAcQqBGCzh/RStIoO8g0NfnfL2MTJRkxoX
                bfDaUeVPQuYEhg37NZWAJQ9VnMVDxP/VHL496M/QZxkjf5/Efucp2gaD
                X6RS6CXpoY68LsvPVjR0ZSwzz1apAzvN9dlzEheX7ICJBBtuA6G3LQpz
                W5hOA2hzCTMjJPJ8LbqF6dsV6DoBQzgul0sGIcGOYl7OyQdXfZ57relS
                Qageu+ipAdTTJ25AsRTAoub8ONGcLmqrAmRLKBP1dfwhYB4N7knNnulq
                QxA+Uk1ihz0=";
    "." initial-key 257 3 8 "AwEAAaz/tAm8yTn4Mfeh5eyI96WSVexTBAvkMgJzkKTOiW1vkIbzxeF3
                +/4RgWOq7HrxRixHlFlExOLAJr5emLvN7SWXgnLh4+B5xQlNVz8Og8kv
                ArMtNROxVQuCaSnIDdD5LKyWbRd2n9WGe2R8PzgCmr3EgVLrjyBxWezF
                0jLHwVN8efS3rCj/EWgvIWgb9tarpVUDK/b58Da+sqqls3eNbuv7pr+e
                oZG+SrDK6nWeL3c6H5Apxz7LjVc1uTIdsIXxuOLYA4/ilBmSVIzuDWfd
                RUfhHdY6+cn8HFRm+2hM8AnXGXws9555KrUB5qihylGa8subX2Nn6UwN
                R1AkUTV74bU=";
};
zone "." IN {
    type hint;
    file "named.ca";
};
zone "localhost.localdomain" IN {
    type master;
    file "named.localhost";
    allow-update {
        "none";
    };
};
zone "localhost" IN {
    type master;
    file "named.localhost";
    allow-update {
        "none";
    };
};
zone "1.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.0.ip6.arpa" IN {
    type master;
    file "named.loopback";
    allow-update {
        "none";
    };
};
zone "1.0.0.127.in-addr.arpa" IN {
    type master;
    file "named.loopback";
    allow-update {
        "none";
    };
};
zone "0.in-addr.arpa" IN {
    type master;
    file "named.empty";
    allow-update {
        "none";
    };
};
"""


def test_config_no_data():
    with pytest.raises(SkipComponent):
        NamedCheckconf(context_wrap(""))


def test_config_dnssec():
    dnssec_disabled = NamedCheckconf(context_wrap(CONFIG_DNSSEC_DISABLED))
    assert dnssec_disabled.is_dnssec_disabled
    assert dnssec_disabled.dnssec_line == "dnssec-enable no;"
    assert dnssec_disabled.disable_algorithms == {}
    assert dnssec_disabled.disable_ds_digests == {}

    dnssec_enabled = NamedCheckconf(context_wrap(CONFIG_DNSSEC_ENABLED))
    assert not dnssec_enabled.is_dnssec_disabled
    assert dnssec_enabled.dnssec_line is None
    assert dnssec_enabled.disable_algorithms == {}
    assert dnssec_enabled.disable_ds_digests == {}

    # dnssec line is not preset - dnssec is enabled by default
    dnssec_default = NamedCheckconf(context_wrap(CONFIG_DNSSEC_DEFAULT))
    assert not dnssec_default.is_dnssec_disabled
    assert dnssec_default.dnssec_line is None
    assert dnssec_default.disable_algorithms == {}
    assert dnssec_default.disable_ds_digests == {}


def test_config_disabled_sections():
    disabled_sections = NamedCheckconf(context_wrap(CONFIG_DISABLED_SECTIONS))
    assert not disabled_sections.is_dnssec_disabled
    assert disabled_sections.dnssec_line is None
    assert disabled_sections.disable_algorithms == {".": ["RSAMD5", "DSA"]}
    assert disabled_sections.disable_ds_digests == {".": ["GOST"]}


def test_doc_examples():
    env = {
        "named_checkconf": NamedCheckconf(context_wrap(CONFIG_DISABLED_SECTIONS)),
    }
    failed, total = doctest.testmod(named_checkconf, globs=env)
    assert failed == 0



import doctest

from insights.parsers import mssql_api_assessment
from insights.parsers.mssql_api_assessment import MssqlApiAssessment, ContainerMssqlApiAssessment
from insights.tests import context_wrap

API_OUTPUT = """
[
  {
    "Timestamp": "2021-05-05T21:51:55.2317511-04:00",
    "Severity": "Information",
    "TargetType": "Server",
    "TargetName": "ceph4-mon",
    "TargetPath": "Server[@Name='ceph4-mon']",
    "CheckId": "TF174",
    "CheckName": "TF 174 increases plan cache bucket count",
    "Message": "Enable trace flag 174 to increase plan cache bucket count",
    "RulesetName": "Microsoft ruleset",
    "RulesetVersion": "1.0.305",
    "HelpLink": "https://docs.microsoft.com/sql/t-sql/database-console-commands/dbcc-traceon-trace-flags-transact-sql"
  },
  {
    "Timestamp": "2021-05-05T21:51:55.2323431-04:00",
    "Severity": "Information",
    "TargetType": "Server",
    "TargetName": "ceph4-mon",
    "TargetPath": "Server[@Name='ceph4-mon']",
    "CheckId": "TF834",
    "CheckName": "TF 834 enables large-page allocations",
    "Message": "Enable trace flag 834 to use large-page allocations to improve analytical and data warehousing workloads",
    "RulesetName": "Microsoft ruleset",
    "RulesetVersion": "1.0.305",
    "HelpLink": "https://support.microsoft.com/kb/3210239"
  }
]
""".strip()


def test_mssql_api_assessment():
    ret = MssqlApiAssessment(context_wrap(API_OUTPUT))
    assert ret[0]["Severity"] == "Information"
    assert ret[0]["TargetName"] == "ceph4-mon"
    assert ret[0]["CheckId"] == "TF174"
    assert ret[0]["Message"] == "Enable trace flag 174 to increase plan cache bucket count"


def test_mssql_api_assessment_container():
    container_ret = ContainerMssqlApiAssessment(
        context_wrap(
            API_OUTPUT,
            container_id='2869b4e2541c',
            image='registry.access.redhat.com/ubi8/nginx-120',
            engine='podman',
        )
    )
    assert container_ret.container_id == "2869b4e2541c"
    assert container_ret[0]["Severity"] == "Information"
    assert container_ret[0]["TargetName"] == "ceph4-mon"
    assert container_ret[0]["CheckId"] == "TF174"
    assert container_ret[0]["Message"] == "Enable trace flag 174 to increase plan cache bucket count"


def test_mssql_api_assessment_doc_examples():
    env = {
        'mssql_api_assessment_output': MssqlApiAssessment(context_wrap(API_OUTPUT)),
        'container_mssql_api_assessment_output': ContainerMssqlApiAssessment(
            context_wrap(
                API_OUTPUT,
                container_id='2869b4e2541c',
                image='registry.access.redhat.com/ubi8/nginx-120',
                engine='podman',
            )
        )
    }
    failed, total = doctest.testmod(mssql_api_assessment, globs=env)
    assert failed == 0



"""
SoftNetStats - file ``/proc/net/softnet_stat``
==============================================
"""

from .. import Parser, parser
from insights.specs import Specs


@parser(Specs.softnet_stat)
class SoftNetStats(Parser):
    """
    Parses ``/proc/net/softnet_stat`` file contains


    This parser parses the stats from network devices. These stats includes events
    per cpu(in row), number of packets processed i.e packet_process (first column),
    number of packet drops packet_drops (second column), time squeeze i.e net_rx_action
    performed time_squeeze(third column), cpu collision i.e collision occur while obtaining
    device lock while transmitting cpu_collision packets (eighth column), received_rps
    number of times cpu woken up received_rps (ninth column), number of times reached flow
    limit count flow_limit_count (tenth column), backlog status (eleventh column),
    core id (twelfth column).

    The typical contents of ``cat /proc/net/softnet_stat`` file is as below::

        00358fe3 00006283 00000000 00000000 00000000 00000000 00000000 00000000 00000000 000855fc 00000000 00000000 00000000
        00953d1a 00000446 00000000 00000000 00000000 00000000 00000000 00000000 00000000 008eeb9a 00000000 00000000 00000001
        02600138 00004bc7 00000000 00000000 00000000 00000000 00000000 00000000 00000000 02328493 00000000 00000000 00000002
        02883c47 00007e2e 00000000 00000000 00000000 00000000 00000000 00000000 00000000 02280d49 00000000 00000000 00000003
        01a35c9d 0002db94 00000001 00000000 00000000 00000000 00000000 00000000 00000000 008ee93a 00000000 00000000 00000004

        Column-01: packet_process: Packet processed by each CPU.
        Column-02: packet_drop: Packets dropped.
        Column-03: time_squeeze: net_rx_action.
        Column-09: cpu_collision: collision occur while obtaining device lock while transmitting.
        Column-10: received_rps: number of times cpu woken up received_rps.
        Column-11: flow_limit_count: number of times reached flow limit count.
        Column-12: softnet_backlog_len: Backlog status
        Column-13: index: core id owning this softnet_data

    For the proc file format details,
    See:
    * Linux 2.6.39 https://elixir.bootlin.com/linux/v2.6.39/source/net/core/dev.c#L4086
    * * example : "6dcad223 00000000 00000001 00000000 00000000 00000000 00000000 00000000 00000000 00000000"
    * Linux 4.18 https://elixir.bootlin.com/linux/v4.18/source/net/core/net-procfs.c#L162
    * * example : "00022be3 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000 00000000"
    * Linux 5.14 https://elixir.bootlin.com/linux/v5.14/source/net/core/net-procfs.c#L169
    * * example : "00358fe3 00006283 00000000 00000000 00000000 00000000 00000000 00000000 00000000 000855fc 00000000 00000000 00000000"

    .. Note::
                There is minimal documentation about these fields in the file, columns are not labeled and
                could change between the kernel releases. This format is compatible for RHEL-6,
                RHEL-7, RHEL-8 and RHEL-9 releases. Also it is unlikely that the positions of those values will
                change in short term.

    Sample softnet_data::

         00358fe3 00006283 00000000 00000000 00000000 00000000 00000000 00000000 00000000 000855fc 00000000 00000000 00000000
         00953d1a 00000446 00000000 00000000 00000000 00000000 00000000 00000000 00000000 008eeb9a 00000000 00000000 00000001
         02600138 00004bc7 00000000 00000000 00000000 00000000 00000000 00000000 00000000 02328493 00000000 00000000 00000002
         02883c47 00007e2e 00000000 00000000 00000000 00000000 00000000 00000000 00000000 02280d49 00000000 00000000 00000003
         01a35c9d 0002db94 00000001 00000000 00000000 00000000 00000000 00000000 00000000 008ee93a 00000000 00000000 00000004

    Examples:
        >>> softnet_obj.is_packet_drops
        True
        >>> softnet_obj.cpu_instances
        5
        >>> softnet_obj.per_cpu_nstat('packet_drops')
        [25219, 1094, 19399, 32302, 187284]
    """

    def __init__(self, context):

        """ Total number of CPUs listed in the stats records"""
        self.cpu_instances = 0

        """ List of network stats per cpu instace """
        self.cpu_nstats = []
        super(SoftNetStats, self).__init__(context)

    def parse_content(self, contents):
        for line in contents:
            stats = [int(st, 16) for st in line.split(None)]
            nstats = {
                    'packet_process': stats[0],
                    'packet_drops': stats[1],
                    'time_squeeze': stats[2],
                    'cpu_collision': stats[8],
                    'received_rps': None,
                    'flow_limit_count': None,
                    'softnet_backlog_len': None,
                    'index': None,
            }

            # RHEL 6,7
            if len(stats) >= 10:
                nstats.update({'received_rps': stats[9]})

            # RHEL 8
            if len(stats) >= 11:
                nstats.update({'flow_limit_count': stats[10]})

            # RHEL 9
            if len(stats) >= 13:
                nstats.update({
                    'softnet_backlog_len': stats[11],
                    'index': stats[12]
                })

            self.cpu_nstats.append(nstats)

        self.cpu_instances = len(self.cpu_nstats)

    def per_cpu_nstat(self, key):
        """
        Get network stats per column for all cpu.

        Arguments:
            (str): Column name for eg. packet_drops or received_rps.

        Returns:
            (list): Column states per cpu.
        """
        if not self.cpu_nstats:
            return []

        if key not in self.cpu_nstats[0]:
            return []

        return [cpu[key] for cpu in self.cpu_nstats]

    @property
    def is_packet_drops(self):
        """
        It will check for if there is packet drop occurred on any cpu.

        Arguments:
            None: No input argument for the function

        Returns:
            (bool): It will return True if observed packet drops.
        """
        return any(cpu["packet_drops"] > 0 for cpu in self.cpu_nstats)



"""
Disk Usage
==========

Combiners for gathering information from du parsers.
"""

from insights import combiner
from insights.parsers.du import DiskUsageDir


@combiner(DiskUsageDir)
class DiskUsageDirs(dict):
    """
    Combiner for the :class:`insights.parsers.du.DiskUsageDir` parser.

    The parser is multioutput, one parser instance for each directory disk
    usage. This combiner puts all of them back together and presents them as a
    dict where the keys are the directory names and the space usage are the
    values.

    Sample input data for du commands as parsed by the parsers::

        # Output of the command:
        # /bin/du -s -k /var/log
        553500	/var/log

        # Output of the command:
        # /bin/du -s -k /var/lib/pgsql
        519228	/var/lib/pgsql

    Examples:
        >>> type(disk_usage_dirs)
        <class 'insights.combiners.du.DiskUsageDirs'>
        >>> sorted(disk_usage_dirs.keys())
        ['/var/lib/pgsql', '/var/log']
        >>> disk_usage_dirs['/var/lib/pgsql']
        519228
    """
    def __init__(self, du_dirs):
        super(DiskUsageDirs, self).__init__()
        for du in du_dirs:
            self.update(du)



# -*- coding: utf-8 -*-
"""
NssRhel7 - file ``/etc/pki/nss-legacy/nss-rhel7.config``
========================================================
"""

from insights import parser, SysconfigOptions
from insights.specs import Specs


@parser(Specs.nss_rhel7)
class NssRhel7(SysconfigOptions):
    """
    This parser reads the ``/etc/pki/nss-legacy/nss-rhel7.config``
    file. It uses the ``SysconfigOptions`` parser class to convert the file into
    a dictionary of options. It also provides the ``config`` property as a helper
    to retrieve the ``config`` variable.

    Attributes:
        config (union[str, None]): The value of the ``config`` variable if it exists, else None.

    Sample Input::

        # To re-enable legacy algorithms, edit this file
        # Note that the last empty line in this file must be preserved
        library=
        name=Policy
        NSS=flags=policyOnly,moduleDB
        config="disallow=MD5:RC4 allow=DH-MIN=1023:DSA-MIN=1023:RSA-MIN=1023:TLS-VERSION-MIN=tls1.0"


    Examples:
        >>> 'config' in nss_rhel7
        True
        >>> nss_rhel7.config
        'disallow=MD5:RC4 allow=DH-MIN=1023:DSA-MIN=1023:RSA-MIN=1023:TLS-VERSION-MIN=tls1.0'
    """

    @property
    def config(self):
        return self.data.get("config", None)



"""
DockerInspect - Command ``docker inspect --type={TYPE}``
========================================================

This module parses the output of the ``docker inspect`` command.  This uses
the ``core.marshalling.unmarshal`` function to parse the JSON output from the
commands.  The parsed data can be accessed a dictionary via the object itself.

"""
from insights.core import CommandParser
from insights.core.exceptions import SkipComponent
from insights.core.marshalling import unmarshal
from insights.core.plugins import parser
from insights.specs import Specs
from insights.util import deprecated


class DockerInspect(CommandParser, dict):
    """
    .. warning::
        This class is deprecated, please use
        :py:class:`insights.parsers.containers_inspect.ContainersInspect` instead.

    Parse the output of command "docker inspect --type=image" and "docker
    inspect --type=container".  The output of these two commands is formatted
    as JSON, so "json.loads" is an option to parse the output in the future.

    Raises:
        SkipComponent: If content is not provided
    """
    def __init__(self, *args, **kwargs):
        deprecated(
            DockerInspect,
            "Please use the :class:`insights.parsers.containers_inspect.ContainersInspect` instead.",
            "3.2.25"
        )
        super(DockerInspect, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        if not content:
            raise SkipComponent

        content = "\n".join(list(content))
        try:
            inspect_data = unmarshal(content)
            self.update(inspect_data[0])
        except:
            raise SkipComponent

    @property
    def data(self):
        return self


@parser(Specs.docker_image_inspect)
class DockerInspectImage(DockerInspect):
    """
    Parse docker image inspect output using the DockerInspect parser class.

    Sample input::

        [
            {
                "Id": "882ab98aae5394aebe91fe6d8a4297fa0387c3cfd421b2d892bddf218ac373b2",
                "RepoTags": [
                    "rhel7_imagemagick:latest"
                ],
                "RepoDigests": [],
        ...

    Examples:

        >>> image['Id'] == '882ab98aae5394aebe91fe6d8a4297fa0387c3cfd421b2d892bddf218ac373b2'
        True
        >>> image['RepoTags'][0] == 'rhel7_imagemagick:latest'
        True

    """
    pass


@parser(Specs.docker_container_inspect)
class DockerInspectContainer(DockerInspect):
    """
    .. warning::
        This parser is deprecated, please use
        :py:class:`insights.parsers.containers_inspect.ContainersInspect` instead.

    Parse docker container inspect output using the DockerInspect parser class.

    Sample input::

        [
            {
                "Id": "97d7cd1a5d8fd7730e83bb61ecbc993742438e966ac5c11910776b5d53f4ae07",
                "Created": "2016-06-23T05:12:25.433469799Z",
                "Path": "/bin/bash",
                "Args": [],
                "Name": "/hehe2",
                "State": {
                    "Status": "running",
                    "Running": true,
                    "Paused": false,
                    "Restarting": false,
            ...

    Examples:

        >>> container['Id'] == '97d7cd1a5d8fd7730e83bb61ecbc993742438e966ac5c11910776b5d53f4ae07'
        True
        >>> container['Name'] == '/hehe2'
        True
        >>> container.get('State').get('Paused') # sub-dictionaries
        False

    """
    def __init__(self, *args, **kwargs):
        deprecated(
            DockerInspectContainer,
            "Please use the :class:`insights.parsers.containers_inspect.ContainersInspect` instead.",
            "3.2.25"
        )
        super(DockerInspectContainer, self).__init__(*args, **kwargs)



"""
GRUB configuration files
========================

This parser reads the configuration of the GRand Unified Bootloader, versions
1 or 2.

This is currently a fairly simple parsing process.  Data read from the file
is put into roughly three categories:

* **configs**: lines read from the file that aren't boot options (i.e.
  excluding lines that go in the *title* and *menuentry* sections).  These
  are split into pairs on the first '=' sign.
* **title**: (GRUB v1 only) lines prefixed by the word 'title'.  All following
  lines up to the next title line are folded together.
* **menuentry**: (GRUB v2 only) lines prefixed by the word 'menuentry'.  All
  following lines up to the line starting with '}' are treated as part of one
  menu entry.

Each of these categories is (currently) stored as a simple list of tuples.

* For the **configs** dict, the key-value pairs based on the line, split
  on the first '=' character.  If nothing is found after the '=' character,
  then the value is `''`.

* For the **title** and **menuentry** list, dict of each boot entry is stored.

  * The items will be key-value pairs, e.g. ``load_video`` will be stored
    as ``{'load_video': ''}`` and ``set root='hd0,msdos1'`` will be stored
    as ``{'set': "root='hd0,msdos1'"}``.

.. note::
    For GRUB version 2, all lines between the ``if`` and ``fi`` will be ignored
    due to we cannot analyze the result of the bash conditions.

Parsers provided by this module are:

Grub1Config - file ``/boot/grub.conf``
--------------------------------------

Grub1EFIConfig - file ``/boot/efi/EFI/redhat/grub.conf``
--------------------------------------------------------

Grub2Config - file ``/boot/grub/grub2.cfg``
-------------------------------------------

Grub2EFIConfig - file ``/boot/efi/EFI/redhat/grub.cfg``
-------------------------------------------------------

BootLoaderEntries - file ``/boot/loader/entries/*.conf``
--------------------------------------------------------
"""
from insights.components.rhel_version import IsRhel6, IsRhel7, IsRhel8, IsRhel9
from insights.core import Parser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsers import get_active_lines
from insights.specs import Specs


class BootEntry(dict):
    """
    An object representing a boot entry in the Grub Configuration.

    Attributes:
        name (str): Name of the boot entry
        cmdline (str): Cmdline of the boot entry
    """
    def __init__(self, data={}):
        self.update(data)
        self.name = data.get('name', '')
        self.cmdline = data.get('cmdline', '')


class GrubConfig(Parser, dict):
    """
    Parser for configuration for both GRUB versions 1 and 2.
    """

    def parse_content(self, content):
        """
        Parse grub configuration file to create a dict with this structure::

            {
                "configs": {
                    name1: [val1, val2, ...]
                    name2: [val],
                    ...
                },
                "title": [
                    {title: name1, kernel: [val], ...},
                    {title: name2, module: [val1, val2], ...},
                ],
                "menuentry": [
                    {menuentry: name1, insmod: [val1, val2], ...},
                    {menuentry: name2, linux16: [val], ...},
                ],
            }

        """
        def _skip_script(line):
            # Skip the script in Grub2
            if line.startswith('if'):
                _skip_script.if_idx += 1
            elif line.startswith('fi'):
                _skip_script.if_idx -= 1
            return _skip_script.if_idx

        def _parser_line(line):
            sp = [i.strip() for i in line.split(None, 1)]
            val = sp[1] if len(sp) > 1 else ''
            if sp[0] not in entry:
                entry[sp[0]] = []
            entry[sp[0]].append(val)
            # Handle the cmdline
            if sp[0].startswith(('kernel', 'linux')):
                b_entry['cmdline'] = val

        def _parser_entry_line(line):
            if line.startswith('title '):
                name = line.split('title', 1)[1].strip()
                b_entry['name'] = entry['title'] = name
            else:
                sp = line.split('menuentry', 1)[1].split('{', 1)
                name = sp[0].strip()
                if not name:
                    raise ParseException("Cannot parse menuentry line: {0}".format(line_raw))
                b_entry['name'] = entry['menuentry'] = name
                # More things after the {
                if len(sp) > 1 and sp[1]:
                    _parser_line(sp[1])

        self.configs = {}
        self._boot_entries = []
        self.entries = []
        b_entry = {}
        entry = {}
        _skip_script.if_idx = 0
        for line_raw in get_active_lines(content):
            # Handle of lines {
            # Remove the heading and trailing whitespace and curly brackets
            line = line_raw.strip('{} \t')
            if_idx = _skip_script(line)
            if if_idx == 0 and line and not line.startswith('fi'):
                # Handle the title / menuentry {
                if line.startswith(('title ', 'menuentry')):
                    self.entries.append(entry) if entry else None
                    self._boot_entries.append(BootEntry(b_entry)) if b_entry else None
                    b_entry = {}
                    entry = {}
                    _parser_entry_line(line)
                # } End of title / menuentry handling
                # Lines inside of an entry
                elif entry:
                    _parser_line(line)
                # Lines outside of entries
                else:
                    sep = '=' if '=' in line else None
                    sp = [i.strip() for i in line.split(sep, 1)]
                    if sp[0] not in self.configs:
                        self.configs[sp[0]] = []
                    self.configs[sp[0]].append(sp[1] if len(sp) > 1 else '')
                # } End of outside of entry
            # } End of lines handling
        # Store the last entry
        self.entries.append(entry) if entry else None
        self.boot_entries.append(BootEntry(b_entry)) if b_entry else None

        self.update({'configs': self.configs}) if self.configs else None
        self._is_kdump_iommu_enabled_ = self._is_kdump_iommu_enabled()
        self._kernel_initrds = get_kernel_initrds(self.entries)

    def _is_kdump_iommu_enabled(self):
        for l in self._boot_entries:
            if "intel_iommu=on" in l.get('cmdline', ''):
                return True
        return False

    @property
    def boot_entries(self):
        """
        Get all boot entries in GRUB configuration.

        Returns:
            (list): A list of :class:`insights.parsers.grub_conf.BootEntry`
                    objects for each boot entry in below format:
                    - 'name': "Red Hat Enterprise Linux Server"
                    - 'cmdline': "kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet"
        """

        return self._boot_entries

    @property
    def is_kdump_iommu_enabled(self):
        """
        Does any kernel have 'intel_iommu=on' set?

        Returns:
            (bool): ``True`` when 'intel_iommu=on' is set, otherwise returns ``False``
        """
        return self._is_kdump_iommu_enabled_

    @property
    def kernel_initrds(self):
        """
        Get the `kernel` and `initrd` files referenced in GRUB configuration files

        Returns:
            (dict): Returns a dict of the `kernel` and `initrd` files referenced
                    in GRUB configuration files
        """
        return self._kernel_initrds


@parser(Specs.grub_conf, [IsRhel6, IsRhel7])
class Grub1Config(GrubConfig):
    """
    Parser for configuration for GRUB version 1.

    Examples:
        >>> grub1_content = '''
        ... default=0
        ... timeout=0
        ... splashimage=(hd0,0)/grub/splash.xpm.gz
        ... hiddenmenu
        ... title Red Hat Enterprise Linux Server (2.6.32-431.17.1.el6.x86_64)
        ...     kernel /vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M rhgb quiet
        ... title Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)
        ...     kernel /vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet
        ... '''.strip()

        >>> grub1_config.configs.get('default')
        ['0']
        >>> grub1_config.configs.get('hiddenmenu')
        ['']
        >>> grub1_config['title'][0]['kernel']
        ['/vmlinuz-2.6.32-431.17.1.el6.x86_64 crashkernel=128M rhgb quiet']
        >>> grub1_config.entries[1]['title']
        'Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)'
        >>> grub1_config.boot_entries[1].name
        'Red Hat Enterprise Linux Server (2.6.32-431.11.2.el6.x86_64)'
        >>> grub1_config.boot_entries[1].cmdline
        '/vmlinuz-2.6.32-431.11.2.el6.x86_64 crashkernel=128M rhgb quiet'
        >>> grub1_config.is_kdump_iommu_enabled
        False
        >>> grub1_config.kernel_initrds['grub_kernels']
        ['vmlinuz-2.6.32-431.17.1.el6.x86_64', 'vmlinuz-2.6.32-431.11.2.el6.x86_64']
    """

    def __init__(self, *args, **kwargs):
        super(Grub1Config, self).__init__(*args, **kwargs)
        self._version = 1
        self._efi = False
        self.update({'title': self.entries})

    def get_current_title(self):
        """
        Get the current default title from the ``default`` option in the
        main configuration. (GRUB v1 only)

        Returns:
            list: A list of dict contains all settings of the default boot entry:
                  - {title: name1, kernel: [val], ...},
        """
        # if no 'default' in grub.conf, set default to 0
        idx = '0'
        for k, v in self.configs.items():
            if k == 'default':
                idx = v[-1]
                break

        if idx.isdigit():
            idx = int(idx)
            if len(self.entries) > idx:
                return self.entries[idx]

        return None


@parser(Specs.grub_efi_conf, [IsRhel6, IsRhel7])
class Grub1EFIConfig(Grub1Config):
    """
    Parses grub v1 configuration for EFI-based systems
    Content of grub-efi.conf is the same as grub.conf
    """
    def __init__(self, *args, **kwargs):
        super(Grub1EFIConfig, self).__init__(*args, **kwargs)
        self._efi = True


@parser(Specs.grub2_cfg, [IsRhel6, IsRhel7, IsRhel8, IsRhel9])
class Grub2Config(GrubConfig):
    """
    Parser for configuration for GRUB version 2.

    Examples:
        >>> grub2_content = '''
        ... ### BEGIN /etc/grub.d/00_header ###
        ... set pager=1
        ... /
        ... if [ -s $prefix/grubenv ]; then
        ...   load_env
        ... fi
        ... #[...]
        ... if [ x"${feature_menuentry_id}" = xy ]; then
        ...   menuentry_id_option="--id"
        ... else
        ...   menuentry_id_option=""
        ... fi
        ... #[...]
        ... ### BEGIN /etc/grub.d/10_linux ###
        ... menuentry 'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c' {
        ...     load_video
        ...     set gfxpayload=keep
        ...     insmod gzio
        ...     insmod part_msdos
        ...     insmod ext2
        ...     set root='hd0,msdos1'
        ...     if [ x$feature_platform_search_hint = xy ]; then
        ...       search --no-floppy --fs-uuid --set=root --hint-bios=hd0,msdos1 --hint-efi=hd0,msdos1 --hint-baremetal=ahci0,msdos1 --hint='hd0,msdos1'  1184ab74-77b5-4cfa-81d3-fb87b0457577
        ...     else
        ...       search --no-floppy --fs-uuid --set=root 1184ab74-77b5-4cfa-81d3-fb87b0457577
        ...     fi
        ...     linux16 /vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8
        ...     initrd16 /initramfs-3.10.0-327.36.3.el7.x86_64.img
        ... }
        ... '''.strip()

        >>> grub2_config['configs']
        {'set pager': ['1'], '/': ['']}
        >>> grub2_config.entries[0]['menuentry']
        "'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c'"
        >>> grub2_config['menuentry'][0]['insmod']
        ['gzio', 'part_msdos', 'ext2']
        >>> grub2_config.boot_entries[0].name
        "'Red Hat Enterprise Linux Workstation (3.10.0-327.36.3.el7.x86_64) 7.2 (Maipo)' $menuentry_id_option 'gnulinux-3.10.0-123.13.2.el7.x86_64-advanced-fbff9f50-62c3-484e-bca5-d53f672cda7c'"
        >>> grub2_config.boot_entries[0].cmdline
        '/vmlinuz-3.10.0-327.36.3.el7.x86_64 root=/dev/RHEL7CSB/Root ro rd.lvm.lv=RHEL7CSB/Root rd.luks.uuid=luks-96c66446-77fd-4431-9508-f6912bd84194 crashkernel=128M@16M rd.lvm.lv=RHEL7CSB/Swap vconsole.font=latarcyrheb-sun16 rhgb quiet LANG=en_GB.utf8'
        >>> grub2_config.kernel_initrds['grub_kernels'][0]
        'vmlinuz-3.10.0-327.36.3.el7.x86_64'
        >>> grub2_config.is_kdump_iommu_enabled
        False
    """
    def __init__(self, *args, **kwargs):
        super(Grub2Config, self).__init__(*args, **kwargs)
        self._version = 2
        self._efi = False
        self.update({'menuentry': self.entries})


@parser(Specs.grub2_efi_cfg, [IsRhel6, IsRhel7, IsRhel8])
class Grub2EFIConfig(Grub2Config):
    """Parses grub2 configuration for EFI-based systems"""
    def __init__(self, *args, **kwargs):
        super(Grub2EFIConfig, self).__init__(*args, **kwargs)
        self._version = 2
        self._efi = True


@parser(Specs.boot_loader_entries, [IsRhel8, IsRhel9])
class BootLoaderEntries(Parser, dict):
    """
    Parses the ``/boot/loader/entries/*.conf`` files.

    Attributes:
        title(str): the name of the boot entry
        cmdline(str): the cmdline of the saved boot entry

    Raises:
        SkipComponent: when input content is empty or no useful data.
    """
    def parse_content(self, content):
        """
        Parses the ``/boot/loader/entries/*.conf`` files.
        """
        if not content:
            raise SkipComponent()

        self.entry = {}
        self.title = ''
        self.cmdline = ''
        for line in content:
            key, value = [i.strip() for i in line.split(None, 1)]
            self.entry[key] = value
            if key == 'options':
                self.cmdline = value

        if not self.entry:
            raise SkipComponent()

        self.update(self.entry)
        self.title = self.entry.get('title')
        self.is_kdump_iommu_enabled = "intel_iommu=on" in self.cmdline


def get_kernel_initrds(entries):
    kernels = []
    initrds = []
    for entry in entries:
        for name, value in entry.items():
            v = []
            if (name.startswith(('module', 'kernel', 'linux', 'initrd'))):
                if isinstance(value, list):
                    v = [i.split()[0].split('/')[-1] for i in value if i]
                elif value and isinstance(value, str):
                    v = [value.split()[0].split('/')[-1]]
            if v:
                if name.startswith('module'):
                    kernels.extend([i for i in v if 'vmlinuz' in i])
                    initrds.extend([i for i in v if 'initrd' in i or 'initramfs' in i])
                elif (name.startswith(('kernel', 'linux'))):
                    if any('ipxe.lkrn' in i for i in v):
                        # Machine PXE boots the kernel, assume all is ok
                        return {}
                    elif 'xen.gz' not in v:
                        kernels.extend([i for i in v if 'xen.gz' not in i])
                elif name.startswith('initrd') or name.startswith('initrd16'):
                    initrds.extend(v)
    return {'grub_kernels': kernels, 'grub_initrds': initrds}



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import blacklisted
from insights.parsers.blacklisted import BlacklistedSpecs
from insights.tests import context_wrap


SPECS = '{"specs": ["dmesg", "fstab"]}'


def test_blacklisted_doc_examples():
    env = {
        "specs": BlacklistedSpecs(context_wrap(SPECS)),
    }
    failed, total = doctest.testmod(blacklisted, globs=env)
    assert failed == 0


def test_skip():
    with pytest.raises(SkipComponent) as ex:
        BlacklistedSpecs(context_wrap(""))
    assert "Empty output." in str(ex)


def test_blacklisted_specs():
    bs = BlacklistedSpecs(context_wrap(SPECS))
    assert bs.specs[0] == "dmesg"



"""
LsOriginLocalVolumePods - command ``ls -l /var/lib/origin/openshift.local.volumes/pods``
========================================================================================
"""

from insights.specs import Specs
from insights import FileListing, parser, CommandParser


@parser(Specs.ls_origin_local_volumes_pods)
class LsOriginLocalVolumePods(CommandParser, FileListing):
    """
    Class to parse the output of command ``ls -l /var/lib/origin/openshift.local.volumes/pods``.
    See ``FileListing`` class for additional information.

    The typical content is::

        total 0
        drwxr-x---. 5 root root 71 Oct 18 23:20 5946c1f644096161a1242b3de0ee5875
        drwxr-x---. 5 root root 71 Oct 18 23:24 6ea3d5cd-d34e-11e8-a142-001a4a160152
        drwxr-x---. 5 root root 71 Oct 18 23:31 77d6d959-d34f-11e8-a142-001a4a160152
        drwxr-x---. 5 root root 71 Oct 18 23:24 7ad952a0-d34e-11e8-a142-001a4a160152
        drwxr-x---. 5 root root 71 Oct 18 23:24 7b63e8aa-d34e-11e8-a142-001a4a160152


    Examples:

        >>> ls_origin_local_volumes_pods.pods
        ['5946c1f644096161a1242b3de0ee5875', '6ea3d5cd-d34e-11e8-a142-001a4a160152', '77d6d959-d34f-11e8-a142-001a4a160152', '7ad952a0-d34e-11e8-a142-001a4a160152', '7b63e8aa-d34e-11e8-a142-001a4a160152']

    Attributes:
        pods (List): The list of pods uid under the directory /var/lib/origin/openshift.local.volumes/pods
    """

    @property
    def pods(self):
        return self.dirs_of("/var/lib/origin/openshift/local/volumes/pods")



from insights.parsers import logrotate_conf
from insights.parsers.logrotate_conf import LogrotateConf
from insights.tests import context_wrap
import doctest

LOGROTATE_CONF_1 = """
# see "man logrotate" for details
# rotate log files weekly
weekly

# keep 4 weeks worth of backlogs
rotate 4

# create new (empty) log files after rotating old ones
create

# use date as a suffix of the rotated file
dateext

# uncomment this if you want your log files compressed
#compress

# RPM packages drop log rotation information into this directory
include /etc/logrotate.d

# no packages own wtmp and btmp -- we'll rotate them here
/var/log/wtmp {
    monthly
    create 0664 root utmp
        minsize 1M
    rotate 1
}
""".strip()

LOGROTATE_CONF_2 = """
#1343753: cant use * here cause not all logs in this folder will be owned by tomcat
/var/log/candlepin/access.log /var/log/candlepin/audit.log /var/log/candlepin/candlepin.log /var/log/candlepin/error.log {
# logrotate 3.8 requires the su directive,
# where as prior versions do not recognize it.
#LOGROTATE-3.8#    su tomcat tomcat
    copytruncate
    daily
    rotate 52
    compress
    missingok
    create 0644 tomcat tomcat
}
""".strip()

LOGROTATE_CONF_3 = """
/var/log/cron
/var/log/maillog
/var/log/messages
/var/log/secure
/var/log/spooler
{
    sharedscripts
    postrotate
        /bin/kill -HUP `cat /var/run/syslogd.pid 2> /dev/null` 2> /dev/null || true
    endscript
}
""".strip()

LOGROTATE_MAN_PAGE_DOC = """
# sample file
compress

/var/log/messages {
    rotate 5
    weekly
    postrotate
                /sbin/killall -HUP syslogd
    endscript
}

"/var/log/httpd/access.log" /var/log/httpd/error.log {
    rotate 5
    mail www@my.org
    size=100k
    sharedscripts
    postrotate
                /sbin/killall -HUP httpd
    endscript
}

/var/log/news/news.crit
/var/log/news/olds.crit  {
    monthly
    rotate 2
    olddir /var/log/news/old
    missingok
    postrotate
                kill -HUP `cat /var/run/inn.pid`
    endscript
    nocompress
}
""".strip()


LOGROTATE_CONF_4 = """
/var/log/news/olds.crit  {
    monthly
    rotate 2
    olddir /var/log/news/old
    missingok
    prerotate
        export LANG=C
        ACTLOG_RLOG=/var/log/actlog/selfinfo/postrotate
        {
                E=/var/log/actlog.exports/eventlog.1
                C=/var/log/actlog.exports/cpuload.1
                if [ -e ${C}.gz -a -e $E ] ; then
                        E_backup=eventlog.1-`date -r $E +%F.%H%M%S`
                        echo "WARNING: Both ${C}.gz and $E exist ; move eventlog.1 to sysinfo/${E_backup}"
                        mv -f $E /var/log/actlog/sysinfo/${E_backup}
                fi
        } >>${ACTLOG_RLOG} 2>&1
        exit 0
  endscript
    nocompress
}
""".strip()


def test_web_xml_doc_examples():
    env = {
        'log_rt': logrotate_conf.LogrotateConf(context_wrap(LOGROTATE_MAN_PAGE_DOC, path='/etc/logrotate.conf')),
        'log_rt_peg': logrotate_conf.LogRotateConfPEG(context_wrap(LOGROTATE_MAN_PAGE_DOC, path='/etc/logrotate.conf')),
    }
    failed, total = doctest.testmod(logrotate_conf, globs=env)
    assert failed == 0


def test_logrotate_conf_1():
    log_rt = LogrotateConf(context_wrap(LOGROTATE_CONF_1, path='/etc/logrotate.conf'))
    assert 'compress' not in log_rt.options
    assert log_rt['include'] == '/etc/logrotate.d'
    assert log_rt['/var/log/wtmp']['minsize'] == '1M'
    assert log_rt.log_files == ['/var/log/wtmp']
    assert log_rt['/var/log/wtmp']['create'] == '0664 root utmp'


def test_logrotate_conf_2():
    log_rt = LogrotateConf(context_wrap(LOGROTATE_CONF_2, path='/etc/logrotate.conf'))
    assert log_rt.options == []
    assert '/var/log/candlepin/access.log' in log_rt.log_files
    assert log_rt['/var/log/candlepin/access.log']['rotate'] == '52'
    assert log_rt['/var/log/candlepin/error.log']['missingok'] is True
    assert log_rt['/var/log/candlepin/audit.log']['create'] == '0644 tomcat tomcat'


def test_logrotate_conf_3():
    log_rt = LogrotateConf(context_wrap(LOGROTATE_CONF_3, path='/etc/logrotate.conf'))
    assert log_rt.options == []
    assert '/var/log/maillog' in log_rt.log_files
    assert log_rt['/var/log/cron']['sharedscripts'] is True
    assert log_rt['/var/log/messages']['postrotate'] == [
            '/bin/kill -HUP `cat /var/run/syslogd.pid 2> /dev/null` 2> /dev/null || true']


def test_logrotate_conf_4():
    log_rt = LogrotateConf(context_wrap(LOGROTATE_CONF_4, path='/etc/logrotate.d/abc'))
    assert '/var/log/news/olds.crit' in log_rt.log_files
    assert 'mv -f $E /var/log/actlog/sysinfo/${E_backup}' in log_rt['/var/log/news/olds.crit']['prerotate']
    assert '} >>${ACTLOG_RLOG} 2>&1' in log_rt['/var/log/news/olds.crit']['prerotate']
    assert len(log_rt['/var/log/news/olds.crit']['prerotate']) == 12



"""
LsVarLibRsyslog - command ``ls -lZ  /var/lib/rsyslog``
======================================================
"""


from insights.core import CommandParser, FileListing
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.ls_var_lib_rsyslog)
class LsVarLibRsyslog(CommandParser, FileListing):
    """
    Parses output of ``ls -lZ  /var/lib/rsyslog`` command.

    Sample output::

        total 4
        -rw-------. 1 root root system_u:object_r:syslogd_var_lib_t:s0 127 Nov 30 03:40 imjournal.state

    Examples:
        >>> rsyslog_obj.dir_contains('/var/lib/rsyslog', 'imjournal.state')
        True
        >>> imjournal_entry = rsyslog_obj.dir_entry('/var/lib/rsyslog', 'imjournal.state')
        >>> imjournal_entry['se_type']
        'syslogd_var_lib_t'
    """
    pass



"""
CrioConf - files ``/etc/crio/crio.conf`` and ``/etc/crio/crio.conf.d/*``
========================================================================
"""

from insights.core import IniConfigFile
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.crio_conf)
class CrioConf(IniConfigFile):
    """
    The ``CrioConf`` class parses the information in the files
    ``/etc/crio/crio.conf`` and ``/etc/crio/crio.conf.d/*``.
    See the ``IniConfigFile`` class for more information on attributes and methods.

    Sample input data looks like::

        [crio]
        storage_option=[ "overlay.imagestore=/mnt/overlay", "overlay.size=1G", ]

        [crio.runtime]
        hooks_dir = [ "/etc/containers/oci/hooks.d", ]
        selinux = true
        default_sysctls = [
        ]

        test_sysctls = [
        \t
        ]

        default_env = [
            "NSS_SDB_USE_CACHE=no",
        ]

        [crio.network]
        plugin_dirs = [
            "/var/lib/cni/bin",
            "/usr/libexec/cni",
        ]

        [crio.metrics]
        metrics_opt = [ "operations" ]
        enable_metrics = true
        metrics_port = 9537
        metrics_collectors = [
          "operations",
          "operations_latency_microseconds_total",
          "operations_latency_microseconds",
          "operations_errors"
          "containers_oom_total",
          "containers_oom",
          # Drop metrics with excessive label cardinality.
          # "image_pulls_by_digest",
          "image_pulls_layer_size",
          # "image_pulls_by_name",
          # "image_pulls_by_name_skipped",
          # "image_pulls_failures",
          # "image_pulls_successes",
          # "image_layer_reuse",
        ]

    Examples:
        >>> "crio" in crio_conf
        True
        >>> crio_conf.has_option('crio.runtime', 'selinux')
        True
        >>> crio_conf.get('crio.metrics', 'metrics_port')
        '9537'
    """
    pass



import doctest
import pytest

from insights.core.exceptions import SkipComponent
from insights.parsers import gluster_peer_status
from insights.tests import context_wrap

OUTPUT = """
Number of Peers: 1

Hostname: versegluster1.verse.loc
Uuid: 86c0266b-c78c-4d0c-afe7-953dec143530
State: Peer in Cluster (Connected)
""".strip()

OUTPUT_1 = """
Number of Peers: 3

Hostname: versegluster1.verse.loc
Uuid: 86c0266b-c78c-4d0c-afe7-953dec143530
State: Peer in Cluster (Connected)

Hostname: 10.30.32.16
Uuid: 3b4673e3-5e95-4c02-b9bb-2823483e067b
State: Peer in Cluster (Connected)

Hostname: 10.30.32.20
Uuid: 4673e3-5e95-4c02-b9bb-2823483e067bb3
State: Peer in Cluster (Disconnected)
""".strip()

OUTPUT_NG = """
Connection failed. Please check if gluster daemon is operational.
""".strip()


def test_output():
    output = gluster_peer_status.GlusterPeerStatus(context_wrap(OUTPUT_1))
    assert output.status['peers'] == len(output.status.get('hosts', []))
    assert output.status.get('hosts', []) == [
        {'Hostname': 'versegluster1.verse.loc', 'State': 'Peer in Cluster (Connected)', 'Uuid': '86c0266b-c78c-4d0c-afe7-953dec143530'},
        {'Hostname': '10.30.32.16', 'State': 'Peer in Cluster (Connected)', 'Uuid': '3b4673e3-5e95-4c02-b9bb-2823483e067b'},
        {'Hostname': '10.30.32.20', 'State': 'Peer in Cluster (Disconnected)', 'Uuid': '4673e3-5e95-4c02-b9bb-2823483e067bb3'}
    ]


def test_blank_output():
    with pytest.raises(SkipComponent) as e:
        gluster_peer_status.GlusterPeerStatus(context_wrap(""))
    assert "No data." in str(e)


def test_failed_output():
    with pytest.raises(SkipComponent) as e:
        gluster_peer_status.GlusterPeerStatus(context_wrap(OUTPUT_NG))
    assert "No data." in str(e)


def test_documentation():
    failed_count, tests = doctest.testmod(
        gluster_peer_status,
        globs={'output': gluster_peer_status.GlusterPeerStatus(context_wrap(OUTPUT))}
    )
    assert failed_count == 0



from insights.tests import context_wrap
from insights.parsers.rhn_charsets import RHNCharSets


emb_charsets_content = """
 server_encoding
-----------------
 UTF~
(1 row)

 client_encoding
-----------------
 UTF8
(1 row)
"""

ora_charsets_content = """
PARAMETER                  VALUE
---------------------------------
NLS_CHARACTERSET           UTF8
NLS_NCHAR_CHARACTERSET     UTF8
"""


def test_embedded_db():
    result = RHNCharSets(context_wrap(emb_charsets_content))
    assert result.get('server_encoding') == 'UTF~'
    assert result.get('client_encoding') == 'UTF8'


def test_oracle_db():
    result = RHNCharSets(context_wrap(ora_charsets_content))
    assert result.get('NLS_CHARACTERSET') == 'UTF8'
    assert result.get('NLS_NCHAR_CHARACTERSET') == 'UTF8'



import os
import six
import copy
import base64
import tempfile
import pkgutil
import hashlib
import insights.client.apps.ansible
from logging import getLogger
from insights.client.apps.ansible.playbook_verifier.contrib import gnupg
from insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel import yaml
from insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel.yaml.comments import CommentedMap, CommentedSeq
from insights.client.apps.ansible.playbook_verifier.contrib.ruamel_yaml.ruamel.yaml.scalarint import ScalarInt
from insights.client.constants import InsightsConstants as constants

__all__ = ("loadPlaybookYaml", "verify", "PlaybookVerificationError")

SIGKEY = 'insights_signature'
PUBLIC_KEY_FOLDER = pkgutil.get_data(insights.client.apps.ansible.__name__, 'playbook_verifier/public.gpg')    # Update this when we have the key generated
EXCLUDABLE_VARIABLES = ['hosts', 'vars']

logger = getLogger(__name__)

yaml = yaml.YAML(typ='rt')
yaml.indent(mapping=2, sequence=4, offset=2)
yaml.default_flow_style = False
yaml.preserve_quotes = True
yaml.width = 200


class PlaybookVerificationError(Exception):
    """
    Exception raised when playbook verification fails

    Attributes:
        playbook -- stringified playbook yaml from stdin
        message -- explanation of why verification failed
    """

    def __init__(self, message="PLAYBOOK VALIDATION FAILED"):
        self.message = message
        super(PlaybookVerificationError, self).__init__(self.message)

    def __str__(self):
        return self.message


def decodeSignature(encodedSignature):
    try:
        decodedSignature = base64.b64decode(encodedSignature)
        return decodedSignature
    except:
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Error Decoding Signature')


def createSnippetHash(snippet):
    """
    Function that creates and returns a hash of the snippet given to the function.
        output: snippetHash (bytes)
    """
    snippetHash = hashlib.sha256()
    if six.PY2:
        normalizedSnippet = normalizeSnippet(snippet)
        serializedSnippet = str(normalizedSnippet).encode("UTF-8")
    else:
        serializedSnippet = str(snippet).encode("UTF-8")
    snippetHash.update(serializedSnippet)

    return snippetHash.digest()


def getPublicKey(gpg):
    if not PUBLIC_KEY_FOLDER:
        raise PlaybookVerificationError(message="PUBLIC KEY IMPORT ERROR: Public key file not found")

    publicKey = PUBLIC_KEY_FOLDER
    importResults = gpg.import_keys(publicKey)
    if (importResults.count < 1):
        raise PlaybookVerificationError(message="PUBLIC KEY NOT IMPORTED: Public key import failed")

    return importResults


def excludeDynamicElements(snippet):
    if 'insights_signature_exclude' not in snippet['vars']:
        raise PlaybookVerificationError(message='EXCLUDE MISSING: the insights_signature_exclude var does not exist.')

    exclusions = snippet['vars']['insights_signature_exclude'].split(',')

    for element in exclusions:
        element = element.split('/')

        # remove empty strings
        element = [string for string in element if string != '']

        if (len(element) == 1 and element[0] in EXCLUDABLE_VARIABLES and element[0] in snippet.keys()):
            del snippet[element[0]]
        elif (len(element) == 2 and element[0] in EXCLUDABLE_VARIABLES):
            try:
                del snippet[element[0]][element[1]]
            except:
                raise PlaybookVerificationError(message='INVALID FIELD: the variable {0} defined in insights_signature_exclude does not exist.'.format(element))
        else:
            raise PlaybookVerificationError(message='INVALID EXCLUSION: the variable {0} is not a valid exclusion.'.format(element))

    return snippet


def executeVerification(snippet, encodedSignature):
    gpg = gnupg.GPG(gnupghome=constants.insights_core_lib_dir)
    snippetHash = createSnippetHash(snippet)
    decodedSignature = decodeSignature(encodedSignature)

    # load public key
    getPublicKey(gpg)

    fd, fn = tempfile.mkstemp()
    os.write(fd, decodedSignature)
    os.close(fd)

    result = gpg.verify_data(fn, snippetHash)
    os.unlink(fn)

    return result, snippetHash


def verifyPlaybookSnippet(snippet):
    if ('vars' not in snippet.keys()):
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Vars field not found')
    elif (snippet['vars'] is None):
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Empty vars field')
    elif (SIGKEY not in snippet['vars']):
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Signature not found')

    encodedSignature = snippet['vars'][SIGKEY]
    snippetCopy = copy.deepcopy(snippet)

    snippetCopy = excludeDynamicElements(snippetCopy)

    return executeVerification(snippetCopy, encodedSignature)


def getRevocationList():
    """
    Load the list of revoked playbook snippet hashes from the egg

    Returns:
        dictionary of revocation list entries (name, hash)
    """
    try:
        # Import revoked list yaml.  The yaml is structured as a list of lists, so we can reuse the playbook signing and
        # verification code.  There will only ever be one list, so we just grab the first element...
        revoked_playbooks = yaml.load(pkgutil.get_data('insights', 'revoked_playbooks.yaml'))[0]

    except Exception:
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Error loading revocation list')

    # verify the list signature!
    verified, snippetHash = verifyPlaybookSnippet(revoked_playbooks)

    if not verified:
        raise PlaybookVerificationError(message='VERIFICATION FAILED: Revocation list signature invalid')

    revocationList = revoked_playbooks.get('revoked_playbooks', [])
    return revocationList


def verify(playbook, skipVerify=False):
    """
    Verify the signed playbook.

    Input: unverified playbook (dictionary format)
    Output: "verified" playbook (dictionary format)
    Error: Playbook Verification failure / Playbook Signature not found.
    """
    logger.info('Playbook Verification has started')

    if not skipVerify:
        if not playbook:
            raise PlaybookVerificationError(message="PLAYBOOK VERIFICATION FAILURE: Playbook is empty")

        revocationList = getRevocationList()

        for snippet in playbook:
            verified, snippetHash = verifyPlaybookSnippet(snippet)

            if not verified:
                name = snippet.get('name', 'NAME UNAVAILABLE')
                raise PlaybookVerificationError(message="SIGNATURE NOT VALID: Template [name: {0}] has invalid signature".format(name))

            # check if snippetHash is on the revoked list
            for revokedItem in revocationList:
                if snippetHash == bytearray.fromhex(revokedItem['hash']):
                    raise PlaybookVerificationError(message="REVOKED PLAYBOOK: Template is on the revoked list [name: {0}]".format(revokedItem['name']))

    logger.info('All templates successfully validated')
    return playbook


def loadPlaybookYaml(playbook):
    """
    Load playbook yaml using current yaml library implementation
        output: playbook yaml
    """
    try:
        playbookYaml = yaml.load(playbook)
        return playbookYaml
    except:
        raise PlaybookVerificationError(message="PLAYBOOK VERIFICATION FAILURE: Failed to load playbook yaml because yaml is not valid")


def normalizeSnippet(snippet):
    """
    Normalize python2 snippet and get rid of any default unicode values
        output: normalized snippet
    """
    new = CommentedMap()
    for key, value in snippet.iteritems():
        if isinstance(value, CommentedMap):
            new[key] = CommentedMap(normalizeSnippet(value))
        elif isinstance(value, CommentedSeq):
            new_sequence = CommentedSeq()
            for item in value:
                if isinstance(item, six.text_type):
                    new_sequence.append(item.encode('ascii', 'ignore'))
                elif isinstance(item, CommentedMap):
                    new_sequence.append(normalizeSnippet(item))
                else:
                    new_sequence.append(item)
            new[key] = new_sequence
        elif isinstance(value, six.text_type):
            new[key] = value.encode('ascii', 'ignore')
        elif isinstance(value, ScalarInt):
            new[key] = int(value)
        else:
            new[key] = value

    return new



import doctest

from insights.parsers import vdsm_conf
from insights.tests import context_wrap


VDSM_CONF = '''
[vars]
ssl = true
cpu_affinity = 1

[addresses]
management_port = 54321
qq = 345
'''

VDSM_LOGGER_CONF = '''
[loggers]
keys=root,vds,storage,virt,ovirt_hosted_engine_ha,ovirt_hosted_engine_ha_config,IOProcess,devel

[handlers]
keys=console,syslog,logfile

[formatters]
keys=long,simple,none,sysform

[logger_root]
level=DEBUG
handlers=syslog,logfile
propagate=0

[logger_vds]
level=DEBUG
handlers=syslog,logfile
qualname=vds
propagate=0

[logger_storage]
level=DEBUG
handlers=logfile
qualname=storage
propagate=0

[logger_ovirt_hosted_engine_ha]
level=DEBUG
handlers=
qualname=ovirt_hosted_engine_ha
propagate=1

[logger_ovirt_hosted_engine_ha_config]
level=DEBUG
handlers=
qualname=ovirt_hosted_engine_ha.env.config
propagate=0

[logger_IOProcess]
level=DEBUG
handlers=logfile
qualname=IOProcess
propagate=0

[logger_virt]
level=DEBUG
handlers=logfile
qualname=virt
propagate=0

[logger_devel]
level=DEBUG
handlers=logfile
qualname=devel
propagate=0

[handler_syslog]
level=WARN
class=handlers.SysLogHandler
formatter=sysform
args=('/dev/log', handlers.SysLogHandler.LOG_USER)

[handler_logfile]
class=vdsm.logUtils.UserGroupEnforcingHandler
args=('vdsm', 'kvm', '/var/log/vdsm/vdsm.log',)
filters=storage.misc.TracebackRepeatFilter
level=DEBUG
formatter=long

[handler_console]
class: StreamHandler
args: []
formatter: none

[formatter_simple]
format: %(asctime)s:%(levelname)s:%(message)s

[formatter_none]
format: %(message)s

[formatter_long]
format: %(asctime)s %(levelname)-5s (%(threadName)s) [%(name)s] %(message)s (%(module)s:%(lineno)d)
class: vdsm.logUtils.TimezoneFormatter

[formatter_sysform]
format= vdsm %(name)s %(levelname)s %(message)s
datefmt=
'''


def test_vdsm_conf_ini():
    result = vdsm_conf.VDSMConfIni(context_wrap(VDSM_CONF))
    assert sorted(result.sections()) == sorted(['vars', 'addresses'])
    assert result.has_option('vars', 'ssl')
    assert result.getboolean('vars', 'ssl')
    assert result.getint('vars', 'cpu_affinity') == 1
    assert result.getint('addresses', 'management_port') == 54321
    assert result.getint('addresses', 'qq') == 345


def test_vdsm_logger_conf():
    conf = vdsm_conf.VDSMLoggerConf(context_wrap(VDSM_LOGGER_CONF))
    assert len(conf.sections()) == 18
    assert conf.has_option('loggers', 'keys') is True
    assert conf.getboolean('logger_root', 'propagate') is False
    assert conf.get('logger_ovirt_hosted_engine_ha', 'level') == 'DEBUG'
    assert conf.get('formatter_sysform', 'datefmt') == ''
    assert conf.has_option('formatter_long', 'class') is True
    assert conf.items('loggers') == {'keys': 'root,vds,storage,virt,ovirt_hosted_engine_ha,ovirt_hosted_engine_ha_config,IOProcess,devel'}


def test_documentation():
    env = {'conf': vdsm_conf.VDSMConfIni(context_wrap(VDSM_CONF)),
           'vdsm_logger_conf': vdsm_conf.VDSMLoggerConf(context_wrap(VDSM_LOGGER_CONF))}
    failed_count, tests = doctest.testmod(vdsm_conf, globs=env)
    assert failed_count == 0



"""
Custom datasources for containers inspect information
"""
import json
import os

from insights.core.context import HostContext
from insights.core.exceptions import ContentException, SkipComponent
from insights.core.filters import get_filters
from insights.core.plugins import datasource
from insights.core.spec_factory import DatasourceProvider, foreach_execute
from insights.specs import Specs
from insights.specs.datasources.container import running_rhel_containers


@datasource(running_rhel_containers, HostContext)
def running_rhel_containers_id(broker):
    """
    Returns a list of container_id of the running containers.
    """
    containers_info = []
    for container in broker[running_rhel_containers]:
        containers_info.append((container[1], container[2]))
    return containers_info


class LocalSpecs(Specs):
    """ Local specs used only by docker|podman_inspect datasources """

    containers_inspect_data_raw = foreach_execute(running_rhel_containers_id, "/usr/bin/%s inspect %s")
    """ Returns the output of command ``/usr/bin/docker|podman inspect <container ID>`` """


@datasource(LocalSpecs.containers_inspect_data_raw, HostContext)
def containers_inspect_data_datasource(broker):
    """
    This datasource provides the filtered information collected
    from ``/usr/bin/docker|podman inspect <container ID>``.

    .. note::
        The path of the target key is like raw_data[key1][key2][target_key],
        then the filter pattern is like "key1|key2|target_key".
        If the value type of raw_data[key1][key2] is list, although the target_key is in the list,
        the filter pattern must be "key1|key2", this datasource returns the whole value of the list.
        The value of "Id" and "Image" are checked from raw data directly, no need filter these items.

    Typical content of ``/usr/bin/docker|podman inspect <container ID>`` file is::

        [
            {
                "Id": "aeaea3ead52724bb525bb2b5c619d67836250756920f0cb9884431ba53b476d8",
                "Created": "2022-10-21T23:47:24.506159696-04:00",
                "Path": "sleep"
                ...
            }
            ...
        ]

    Returns:
        list: item is JSON string containing filtered information.

    Raises:
        SkipComponent: When the filter/path does not exist or any exception occurs.
    """
    try:
        no_filters = ['Id', 'Image']
        filters = sorted(set(get_filters(Specs.container_inspect_keys)) - set(no_filters))
        if filters:
            total_results = []
            for item in broker[LocalSpecs.containers_inspect_data_raw]:
                raw_data = json.loads(''.join(item.content))[0]
                engine, _, container_id = item.cmd.split(None)
                filter_result = dict(Id=container_id, engine=os.path.basename(engine))
                if 'Image' in raw_data:
                    filter_result['Image'] = raw_data['Image'].split("sha256:")[-1]
                for item in filters:
                    val = raw_data
                    for key in item.split("|"):
                        if key in val:
                            val = val[key]
                        else:
                            break
                    # If the filtered key does not exist, skip it
                    if val == raw_data:
                        continue
                    filter_result[item] = val
                total_results.append(filter_result)
            # Generally the False branch of this condition will never reach, since the required component
            # LocalSpecs.containers_inspect_data_raw can guarantee total_results is not null. However, it's worth
            # leaving this condition as an explicit assertion to avoid unexpected situation.
            if total_results:
                return DatasourceProvider(content=json.dumps(total_results), relative_path='insights_containers/containers_inspect')
    except Exception as e:
        raise ContentException("Unexpected content exception:{e}".format(e=str(e)))
    raise SkipComponent



"""
cpu_online - File ``/sys/devices/system/cpu/online``
====================================================

This parser reads the content of ``/sys/devices/system/cpu/online``.
This file shows the number of online cpu. The format of the content
is string including comma.
"""

from insights.core import CommandParser, ContainerParser
from insights.core.plugins import parser
from insights.specs import Specs


@parser(Specs.container_cpu_online)
class ContainerCpuOnline(ContainerParser, CommandParser):
    """
    Class ``ContainerCpuOnline`` parses the content of the ``/sys/devices/system/cpu/online`` from containers.

    Attributes:
        cpu_online_set (list): It is used to show the list of online cpu.

        cpu_online_number (int): It is used to display the number of online cpu.

    A small sample of the content of this file looks like::

        0,2-4,7

    Examples:
        >>> type(container_cpu_online_info)
        <class 'insights.parsers.cpu_online.ContainerCpuOnline'>
        >>> container_cpu_online_info.container_id
        '2869b4e2541c'
        >>> container_cpu_online_info.image
        'registry.access.redhat.com/ubi8/nginx-120'
        >>> container_cpu_online_info.cpu_online_set
        ['0', '2', '3', '4', '7']
        >>> container_cpu_online_info.cpu_online_number
        5
    """
    def parse_content(self, content):
        self.cpu_online_set = []
        self.cpu_online_number = 0
        values = content[0].strip().split(",")
        for value in values:
            if "-" in value:
                # Parse the value like "2-4"
                start, end = value.split("-")
                self.cpu_online_set.extend([str(i) for i in range(int(start), int(end) + 1)])
            else:
                self.cpu_online_set.append(value)
        self.cpu_online_number = len(self.cpu_online_set)



"""
RedhatRelease - File ``/etc/redhat-release``
============================================

This module provides plugins access to file ``/etc/redhat-release``

Typical content of file ``/etc/redhat-release`` is::

    Red Hat Enterprise Linux Server release 7.2 (Maipo)

This module parses the file contents and stores data in the class
attributes described below.

Examples:
    >>> type(rh_release)
    <class 'insights.parsers.redhat_release.RedhatRelease'>
    >>> rh_release.raw
    'Red Hat Enterprise Linux Server release 7.2 (Maipo)'
    >>> rh_release.major
    7
    >>> rh_release.minor
    2
    >>> rh_release.version
    '7.2'
    >>> rh_release.is_rhel
    True
    >>> rh_release.product
    'Red Hat Enterprise Linux Server'
"""
from insights import Parser, ContainerParser, parser
from insights.specs import Specs


@parser(Specs.redhat_release)
class RedhatRelease(Parser):
    """Parses the content of file ``/etc/redhat-release``

    Attributes:
        is_alpha(bool): True if this is an Alpha release
        is_beta(bool): True if this is a Beta release
        is_centos(bool): True if this release is CentOS
        is_fedora(bool): True if this release is Fedora
        is_rhel(bool): True if this release is Red Hat Enterprise Linux
        major(int): Major release number or None
        minor(int): Minor release number or None
        parsed(dict): Dictionary containing the parsed strings for ``product``, ``version``, and ``code_name``
        raw(string): Unparsed redhat-release string
    """

    def parse_content(self, content):
        self.raw = content[0]
        self.is_beta = False
        self.is_alpha = False
        product, _, version_name = [v.strip() for v in content[0].partition("release")]
        if 'Alpha' in version_name:
            # Red Hat Enterprise Linux release 9 Alpha (Plow)
            version_number, code_name = version_name.split('Alpha', 1)
            self.is_alpha = True
        elif 'Beta' in version_name:
            # Red Hat Enterprise Linux release 8.5 Beta (Ootpa)
            version_number, code_name = version_name.split('Beta', 1)
            self.is_beta = True
        elif '(' in version_name:
            # Red Hat Enterprise Linux Workstation release 6.10(Santiago)
            # Red Hat Enterprise Linux Workstation release 6.10 (Santiago)
            version_number, code_name = version_name.split('(', 1)
        else:
            # Red Hat Enterprise Linux Workstation release 6.10
            version_number = version_name
            code_name = None

        self.parsed = {
            "product": product,
            "version": version_number.strip(),
            "code_name": code_name.strip().strip('()') if code_name is not None else None
        }

        self.is_rhel = 'red hat enterprise linux' in self.parsed['product'].lower()
        self.is_centos = 'centos' in self.parsed['product'].lower()
        self.is_fedora = 'fedora' in self.parsed['product'].lower()

        v_parts = self.parsed['version'].split('.')
        self.major = int(v_parts[0]) if v_parts[0].isdigit() else None
        if len(v_parts) >= 2:
            if '-' in v_parts[1]:
                minor = v_parts[1].split('-')[0]
            else:
                minor = v_parts[1]
            self.minor = int(minor) if minor.isdigit() else None
        else:
            self.minor = None

    @property
    def version(self):
        """string: version of this OS."""
        return self.parsed["version"]

    @property
    def product(self):
        """string: product of this OS."""
        return self.parsed["product"]

    @property
    def code_name(self):
        """string: code name of this OS or None."""
        return self.parsed["code_name"]


@parser(Specs.container_redhat_release)
class ContainerRedhatRelease(ContainerParser, RedhatRelease):
    """
    Parses the content of file ``/etc/redhat-release`` of the running
    containers which are based on RHEL images.
    """
    @property
    def rhel(self):
        """string: alias of ``self.version``"""
        return self.version



"""
Mount Entries
=============

Parsers provided in this module includes:

Mount - command ``/bin/mount``
------------------------------

ProcMounts - file ``/proc/mounts``
----------------------------------

MountInfo - file ``/proc/self/mountinfo``
-----------------------------------------

The ``Mount`` class implements parsing for the ``mount`` command output which looks like::

    /dev/mapper/rootvg-rootlv on / type ext4 (rw,relatime,barrier=1,data=ordered)
    proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)
    /dev/mapper/HostVG-Config on /etc/shadow type ext4 (rw,noatime,seclabel,stripe=256,data=ordered)
    dev/sr0 on /run/media/root/VMware Tools type iso9660 (ro,nosuid,nodev,relatime,uid=0,gid=0,iocharset=utf8,mode=0400,dmode=0500,uhelper=udisks2) [VMware Tools]

The information is stored as a list of :class:`MountEntry` objects.  Each
:class:`MountEntry` object contains attributes for the following information that
are listed in the same order as in the command output:

 * ``mount_source`` - Name of filesystem or the mounted device
 * ``mount_point`` - Name of mount point for filesystem
 * ``mount_type`` - Name of filesystem type
 * ``mount_options`` -  Mount options as ``MountOpts`` object
 * ``mount_label`` - Optional label of this mount entry, empty string by default
 * ``mount_addtlinfo`` - Additional mount information as ``MountAddtlInfo`` object
 * ``mount_clause`` - Full raw string from command output
 * ``filesystem`` - Name of filesystem or the mounted device (Deprecated)

``ProcMounts`` and ``MountInfo`` classes have the similar style as ``mount``.

MountEntry lines are also available in a ``mounts`` property, keyed on the
mount point.
"""
import os

from insights.core import CommandParser
from insights.core.exceptions import ParseException, SkipComponent
from insights.core.plugins import parser
from insights.parsers import get_active_lines, keyword_search, optlist_to_dict
from insights.specs import Specs


class AttributeAsDict(object):
    """
    Set the given key, value pair data as attribute of object.

    The ``in`` operator could be used to determine if an attribute exists before
    direct accessing.
    """

    def __init__(self, data=None):
        data = {} if data is None else data
        for k, v in data.items():
            setattr(self, k, v)

    def __contains__(self, item):
        return item in self.__dict__

    def __getitem__(self, item):
        return self.__dict__[item]

    def __len__(self):
        return len(self.__dict__)

    def get(self, item, default=None):
        return self.__dict__.get(item, default)

    def items(self):
        for k, v in self.__dict__.items():
            yield k, v


class MountOpts(AttributeAsDict):
    """
    An object representing the mount options found in mount or fstab entry as
    attributes accessible via the attribute name as it appears in the command
    output.  For instance, the options ``(rw,dmode=0500)`` may be accessed as
    ``mnt_row_info.rw`` with the value ``True`` and ``mnt_row_info.dmode``
    with the value "0500".
    """
    pass


class MountAddtlInfo(AttributeAsDict):
    """
    An object representing the additional information for an mount entry as
    attributes accessible via the attribute name.
    Object will provides different set of attributes according to different data
    sources.

    Attributes:
        mount_label (str): Label of this mount entry from command ``/bin/mount``
        fs_freq (str): fs_freq of this mount entry from file ``/proc/mounts``
        fs_passno (str): fs_passno of this mount entry from file ``/proc/mounts``
        mount_id (str): Unique identifier of the mount from file ``/proc/self/mountinfo``
        parent_id (str): Unique identifier of the parent mount from file ``/proc/self/mountinfo``
        major_minor (str): Value of st_dev for files on filesystem from file ``/proc/self/mountinfo``
        root (str): Root of the mount within the filesystem from file ``/proc/self/mountinfo``
        optional_fields (str): Zero or more fields of the form "tag[:value]" from file ``/proc/self/mountinfo``
        mount_clause_binmount (str): Full raw string from command ``/bin/mount``
        mount_clause_procmounts (str): Full raw string from file ``/proc/mounts``
        mount_clause_mountinfo (str): Full raw string from file ``/proc/self/mountinfo``

    For instance, the major:minor number ``253:4`` could be accessed as
    ``mnt_row_info.major_minor`` with the value ``253:4``.
    """
    pass


class MountEntry(AttributeAsDict):
    """
    An object representing an mount entry of ``mount`` command, ``/proc/mounts``
    or ``/proc/self/mountinfo`` file. Each entry contains below attributes:

    Attributes:
        mount_source(str): Name of filesystem of mounted device
        mount_point (str): Name of mount point for filesystem
        mount_type (str): Name of filesystem type
        mount_options (MountOpts): Mount options as :class:`MountOpts`
        mount_label (str): Optional label of this mount entry, an empty string by default
        mount_addtlinfo (MountAddtlInfo): Additional mount information as :class:`MountAddtlInfo`
        mount_clause (str): Full raw string from command output
        filesystem (str): Name of filesystem of mounted device (Deprecated, use
                `mount_source` instead)
    """
    pass


class MountedFileSystems(CommandParser):
    """
    Base Class for :class:`Mount`, :class:`ProcMounts` and :class:`MountInfo`.

    Attributes:
        rows (list): List of :class:`MountEntry` objects for each row of the
            content.
        mounts (dict): Dict with the `mount_point` as the key and the
            :class:`MountEntry` objects as the value.

    Raises:
        SkipComponent: When the file is empty.
        ParseException: Raised when any problem parsing the command output.
    """
    def __len__(self):
        return len(self.rows)

    def __iter__(self):
        for row in self.rows:
            yield row

    def __getitem__(self, idx):
        if isinstance(idx, int):
            return self.rows[idx]
        elif isinstance(idx, str):
            return self.mounts[idx]
        else:
            raise TypeError("Mounts can only be indexed by mount string or line number")

    def parse_content(self, content):
        # No content found or file is empty
        if not content:
            raise SkipComponent('Empty content')

        self._parse_mounts(content)

        if '/' not in self.mounts:
            raise ParseException("Input for mount must contain '/' mount point.")

    def get_dir(self, path):
        """
        This finds the most specific mount path that contains the given path,
        by successively removing the directory or file name on the end of
        the path and seeing if that is a mount point.  This will always
        terminate since / is always a mount point.  Strings that are not
        absolute paths will return None.

        Arguments:
            path (str): The path to check.
        Returns:
            MountEntry: The mount point that contains the given path.
        """
        while path != '':
            if path in self.mounts:
                return self.mounts[path]
            path = os.path.split(path)[0]
        return None

    def search(self, **kwargs):
        """
        Returns a list of the mounts (in order) matching the given criteria.
        Keys are searched for directly - see the
        :py:func:`insights.parsers.keyword_search` utility function for more
        details.  If no search parameters are given, no rows are returned.

        Examples:

            >>> mounts.search(filesystem='proc')[0].mount_clause
            'proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)'
            >>> mounts.search(mount_options__contains='seclabel')[0].mount_clause
            '/dev/mapper/HostVG-Config on /etc/shadow type ext4 (rw,noatime,seclabel,stripe=256,data=ordered)'

        Arguments:
            **kwargs (dict): Dictionary of key-value pairs to search for.

        Returns:
            (list): The list of mount points matching the given criteria.
        """
        return keyword_search(self.rows, **kwargs)


@parser(Specs.mount)
class Mount(MountedFileSystems):
    """
    Class of information for all output from ``mount`` command.

    .. note::
        Please refer to its super-class :class:`MountedFileSystems` for more
        details.

    The typical output of ``mount`` command looks like::

        /dev/mapper/rootvg-rootlv on / type ext4 (rw,relatime,barrier=1,data=ordered)
        proc on /proc type proc (rw,nosuid,nodev,noexec,relatime)
        /dev/mapper/HostVG-Config on /etc/shadow type ext4 (rw,noatime,seclabel,stripe=256,data=ordered)
        dev/sr0 on /run/media/root/VMware Tools type iso9660 (ro,nosuid,nodev,relatime,uid=0,gid=0,iocharset=utf8,mode=0400,dmode=0500,uhelper=udisks2) [VMware Tools]

    Examples:

        >>> type(mnt_info)
        <class 'insights.parsers.mount.Mount'>
        >>> len(mnt_info)
        4
        >>> mnt_info[3].filesystem
        'dev/sr0'
        >>> mnt_info[3].mount_label
        '[VMware Tools]'
        >>> mnt_info[3].mount_type
        'iso9660'
        >>> 'ro' in mnt_info[3].mount_options
        True
        >>> mnt_info['/run/media/root/VMware Tools'].filesystem
        'dev/sr0'
        >>> mnt_info['/run/media/root/VMware Tools'].mount_label
        '[VMware Tools]'
        >>> mnt_info['/run/media/root/VMware Tools'].mount_options.ro
        True
    """
    def _parse_mounts(self, content):
        self.rows = []
        self.mounts = {}
        for line in get_active_lines(content):
            mount = {}
            mount['mount_clause'] = line
            # Get the mounted filesystem by checking the ' on '
            line_sp = _customized_split(line, line, sep=' on ')
            mount['mount_source'] = line_sp[0]
            mount['filesystem'] = mount['mount_source']  # compatible for deprecation
            # Get the mounted point by checking the last ' type ' before the last '('
            mnt_pt_sp = _customized_split(raw=line, l=line_sp[1], sep=' (', reverse=True)
            line_sp = _customized_split(raw=line, l=mnt_pt_sp[0], sep=' type ', reverse=True)
            mount['mount_point'] = line_sp[0]
            mount['mount_type'] = line_sp[1].split()[0]
            line_sp = _customized_split(raw=line, l=mnt_pt_sp[1], sep=None, check=False)
            mount['mount_options'] = MountOpts(optlist_to_dict(line_sp[0].strip('()')))
            if len(line_sp) == 2:
                mount['mount_label'] = line_sp[1]

            entry = MountEntry(mount)
            self.rows.append(entry)
            self.mounts[mount['mount_point']] = entry


@parser(Specs.mounts)
class ProcMounts(MountedFileSystems):
    """
    Class to parse the content of ``/proc/mounts`` file.

    This class is required to parse the ``/proc/mounts`` file in addition to the
    ``/bin/mount`` command because it lists the mount points of those process's
    which are not present in the output of the ``/bin/mount`` command.

    .. note::
        Please refer to its super-class :class:`MountedFileSystems` for more
        details.

    The typical content of ``/proc/mounts`` file looks like::

        /dev/mapper/rootvg-rootlv / ext4 rw,relatime,barrier=1,data=ordered 0 0
        proc /proc proc rw,nosuid,nodev,noexec,relatime 0 0
        /dev/mapper/HostVG-Config /etc/shadow ext4 rw,noatime,seclabel,stripe=256,data=ordered 0 0
        dev/sr0 /run/media/root/VMware\040Tools iso9660 ro,nosuid,nodev,relatime,uid=0,gid=0,iocharset=utf8,mode=0400,dmode=0500,uhelper=udisks2 0 0

    Examples:
        >>> type(proc_mnt_info)
        <class 'insights.parsers.mount.ProcMounts'>
        >>> len(proc_mnt_info)
        4
        >>> proc_mnt_info[3].filesystem == 'dev/sr0'
        True
        >>> proc_mnt_info[3].mounted_device == 'dev/sr0'
        True
        >>> proc_mnt_info[3].mounted_device == proc_mnt_info[3].filesystem
        True
        >>> proc_mnt_info[3].mount_type == 'iso9660'
        True
        >>> proc_mnt_info[3].filesystem_type == 'iso9660'
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mount_label == ['0', '0']
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mount_options.ro
        True
        >>> proc_mnt_info['/run/media/root/VMware Tools'].mounted_device == 'dev/sr0'
        True
    """

    def _parse_mounts(self, content):

        self.rows = []
        self.mounts = {}
        for line in get_active_lines(content):
            mount = {}
            mount['mount_clause'] = line
            # Handle the '\040' in `mount_point`, e.g. "VMware\040Tools"
            line_sp = line.encode().decode("unicode-escape")
            line_sp = _customized_split(raw=line, l=line_sp)
            mount['filesystem'] = mount['mounted_device'] = line_sp[0]
            line_sp = _customized_split(raw=line, l=line_sp[1], num=3, reverse=True)
            mount['mount_label'] = line_sp[-2:]
            line_sp = _customized_split(raw=line, l=line_sp[0], reverse=True)
            mount['mount_options'] = MountOpts(optlist_to_dict(line_sp[1]))
            line_sp = _customized_split(raw=line, l=line_sp[0], reverse=True)
            mount['mount_type'] = mount['filesystem_type'] = line_sp[1]
            mount['mount_point'] = line_sp[0]

            entry = MountEntry(mount)
            self.rows.append(entry)
            self.mounts[mount['mount_point']] = entry


@parser(Specs.mountinfo)
class MountInfo(MountedFileSystems):
    """
    Class to parse the content of ``/proc/self/mountinfo`` file.

    .. note::
        Please refer to its super-class :class:`MountedFileSystems` for more details.

    The typical content of ``/proc/self/mountinfo`` file looks like::

        39 0 253:0 / / rw,relatime shared:1 - xfs /dev/mapper/rootvg-lvlocroot rw,attr2,inode64,noquota
        47 39 8:1 / /boot rw,relatime shared:30 - xfs /dev/sda1 rw,attr2,inode64,noquota
        65 39 253:19 / /data rw,relatime shared:44 - ext4 /dev/mapper/vgdata-lvdata rw,data=ordered
        58 39 253:15 / /opt rw,relatime shared:45 - xfs /dev/mapper/rootvg-lvlocopt rw,attr2,inode64,noquota
        61 39 253:17 / /home rw,nosuid,relatime shared:46 - xfs /dev/mapper/rootvg-lvlochome rw,attr2,inode64,noquotao

    Examples:

        >>> type(proc_mountinfo)
        <class 'insights.parsers.mount.MountInfo'>
        >>> len(proc_mountinfo)
        5
        >>> proc_mountinfo[2].mount_source
        '/dev/mapper/vgdata-lvdata'
        >>> proc_mountinfo[2].mount_type
        'ext4'
        >>> proc_mountinfo[2].mount_point
        '/data'
        >>> 'data' in proc_mountinfo[2].mount_options
        True
        >>> proc_mountinfo[2].mount_options.data
        'ordered'
        >>> 'major_minor' in proc_mountinfo[4]
        False
        >>> 'major_minor' in proc_mountinfo[4].mount_addtlinfo
        True
        >>> proc_mountinfo[4].mount_addtlinfo.major_minor
        '253:17'
        >>> proc_mountinfo['/boot'].mount_source
        '/dev/sda1'
    """

    def _parse_mounts(self, content):
        rows = []
        for line in get_active_lines(content):
            mount = {}
            mount['mount_clause'] = line
            line_own_sp = _customized_split(line, line, sep=' - ')
            line_left_sp = _customized_split(line, line_own_sp[0], num=6)
            mount['mount_point'] = line_left_sp[4]
            line_right_sp = _customized_split(line, line_own_sp[1], num=3)
            mount['mount_type'] = line_right_sp[0]
            mount['mount_source'] = line_right_sp[1]
            mount['filesystem'] = mount['mount_source']  # compatible for deprecation
            optional_fields = line_left_sp[5].split()[1] if ' ' in line_left_sp[5] else ''
            mntopt = line_left_sp[5].split()[0] if ' ' in line_left_sp[5] else line_left_sp[5]
            unioned_options = ','.join([mntopt, line_right_sp[2]])
            mount['mount_options'] = MountOpts(optlist_to_dict(unioned_options))
            mount['mount_addtlinfo'] = MountAddtlInfo({
                    'mount_id': line_left_sp[0],
                    'parent_id': line_left_sp[1],
                    'major_minor': line_left_sp[2],
                    'root': line_left_sp[3],
                    'optional_fields': optional_fields,
            })
            entry = MountEntry(mount)
            rows.append(entry)
        self.rows = rows
        self.mounts = dict([mnt['mount_point'], rows[idx]] for idx, mnt in enumerate(rows))


def _customized_split(raw, l, sep=None, num=2, reverse=False, check=True):
    if num >= 2:
        if reverse is False:
            line_sp = l.split(sep, num - 1)
        else:
            line_sp = l.rsplit(sep, num - 1)
        if check and len(line_sp) < num:
            raise ParseException('Unable to parse: "{0}"'.format(raw))
    return line_sp



from insights.parsers.journald_conf import JournaldConf
from insights.tests import context_wrap

# from RHEL 7.3, tweaked
JOURNALD_CONF_1 = """
#  This file is part of systemd.
#
#  systemd is free software; you can redistribute it and/or modify it
#  under the terms of the GNU Lesser General Public License as published by
#  the Free Software Foundation; either version 2.1 of the License, or
#  (at your option) any later version.
#
# Entries in this file show the compile time defaults.
# You can change settings by editing this file.
# Defaults can be restored by simply deleting this file.
#
# See journald.conf(5) for details.

[Journal]
#Storage=auto
Storage=persistent
#Compress=yes
#Seal=yes
Seal=no
#SplitMode=uid
#SyncIntervalSec=5m
#RateLimitInterval=30s
#RateLimitBurst=1000
#SystemMaxUse=
#SystemKeepFree=
#SystemMaxFileSize=
#RuntimeMaxUse=
#RuntimeKeepFree=
#RuntimeMaxFileSize=
#MaxRetentionSec=
#MaxFileSec=1month
#ForwardToSyslog=yes
ForwardToSyslog=no
#ForwardToKMsg=no
#ForwardToConsole=no
#ForwardToWall=yes
#TTYPath=/dev/console
#MaxLevelStore=debug
#MaxLevelSyslog=debug
#MaxLevelKMsg=notice
#MaxLevelConsole=info
#MaxLevelWall=emerg

""".strip()

JOURNALD_CONF_2 = """
#comment
# comment
# comment = comment
# comment = comment = comment
#comment=comment
#comment=comment=comment
option_a=value_a
option_b = value_b
option_c= value_c
option_d =value_d
broken_option_e = value_e = value_2_e
broken_option_f=value_f=value_2_f
broken_option_g
option_h = value_h # some comment
option_i = value_i # this must be accessible, even after all these errors
""".strip()

AUDITD_CONF_PATH = "/etc/audit/auditd.conf"


def test_constructor():
    context = context_wrap(JOURNALD_CONF_1, AUDITD_CONF_PATH)
    result = JournaldConf(context)

    assert "Storage=persistent" in result.active_lines_unparsed
    assert "no" == result.active_settings["Seal"]
    assert "#TTYPath" not in result.active_settings
    assert "no" == result.get_active_setting_value("ForwardToSyslog")

    context = context_wrap(JOURNALD_CONF_2, AUDITD_CONF_PATH)
    result = JournaldConf(context)

    assert "comment" not in result.active_settings
    assert "broken_option_g" not in result.active_settings
    assert "value_i" == result.get_active_setting_value("option_i")


def test_active_lines_unparsed():
    context = context_wrap(JOURNALD_CONF_1, AUDITD_CONF_PATH)
    result = JournaldConf(context)
    test_active_lines = []
    for line in JOURNALD_CONF_1.split("\n"):
        if not line.strip().startswith("#"):
            if line.strip():
                test_active_lines.append(line)
    assert test_active_lines == result.active_lines_unparsed


def build_active_settings_expected():
    active_settings = {}
    for line in JOURNALD_CONF_1.split("\n"):
        if not line.strip().startswith("#"):
            if line.strip():
                try:
                    key, value = line.split("=", 1)
                    key, value = key.strip(), value.strip()
                except:
                    pass
                else:
                    active_settings[key] = value
    return active_settings


def test_active_settings():
    context = context_wrap(JOURNALD_CONF_1, AUDITD_CONF_PATH)
    result = JournaldConf(context)
    active_settings = build_active_settings_expected()
    assert active_settings == result.active_settings


def test_get_active_setting_value():
    context = context_wrap(JOURNALD_CONF_1, AUDITD_CONF_PATH)
    result = JournaldConf(context)
    active_settings = build_active_settings_expected()
    for key, value in active_settings.items():
        assert result.get_active_setting_value(key) == value



"""
This module contains logic for parsing ls output. It attempts to handle
output when selinux is enabled or disabled and also skip "bad" lines.
"""
import six


def parse_path(path):
    """
    Convert possible symbolic link into a source -> target pair.

    Args:
        path (str): The path portion of an ls output line.

    Returns:
        A (path, link) tuple where path is always populated and link is a non
        empty string if the original path is a symoblic link.
    """
    path, _, link = path.partition(" -> ")
    return path, link


def parse_non_selinux(parts):
    """
    Parse part of an ls output line that isn't selinux.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are link
            count, owner, group, and everything else.

    Returns:
        A dict containing links, owner, group, date, and name. If the line
        represented a device, major and minor numbers are included.  Otherwise,
        size is included. If the raw name was a symbolic link, link is
        included.
    """
    links, owner, group, last = parts
    result = {
        "links": int(links),
        "owner": owner,
        "group": group,
    }

    # device numbers only go to 256.
    # If a comma is in the first four characters, the next two elements are
    # major and minor device numbers. Otherwise, the next element is the size.
    if "," in last[:4]:
        major, minor, rest = last.split(None, 2)
        result["major"] = int(major.rstrip(","))
        result["minor"] = int(minor)
    else:
        size, rest = last.split(None, 1)
        result["size"] = int(size)

    # The date part is always 12 characters regardless of content.
    result["date"] = rest[:12]

    # Jump over the date and the following space to get the path part.
    path, link = parse_path(rest[13:])
    result["name"] = path
    if link:
        result["link"] = link

    return result


def parse_selinux(parts):
    """
    Parse part of an ls output line that is selinux.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are owner
            group, selinux info, and the path.

    Returns:
        A dict containing owner, group, se_user, se_role, se_type, se_mls, and
        name. If the raw name was a symbolic link, link is also included.

    """

    owner, group = parts[:2]
    selinux = parts[2].split(":")
    lsel = len(selinux)
    path, link = parse_path(parts[-1])
    result = {
        "owner": owner,
        "group": group,
        "se_user": selinux[0],
        "se_role": selinux[1] if lsel > 1 else None,
        "se_type": selinux[2] if lsel > 2 else None,
        "se_mls": selinux[3] if lsel > 3 else None,
        "name": path
    }
    if link:
        result["link"] = link
    return result


def parse_rhel8_selinux(parts):
    """
    Parse part of an ls output line that is selinux on RHEL8.

    Args:
        parts (list): A four element list of strings representing the initial
            parts of an ls line after the permission bits. The parts are link
            count, owner, group, and everything else

    Returns:
        A dict containing links, owner, group, se_user, se_role, se_type,
        se_mls, size, date, and name. If the raw name was a symbolic link,
        link is also included.

    """

    links, owner, group, last = parts

    selinux = parts[3].split(":")
    lsel = len(selinux)
    selinux, size, last = parts[-1].split(None, 2)
    selinux = selinux.split(":")
    date = last[:12]
    path, link = parse_path(last[13:])
    result = {
        "links": int(links),
        "owner": owner,
        "group": group,
        "se_user": selinux[0],
        "se_role": selinux[1] if lsel > 1 else None,
        "se_type": selinux[2] if lsel > 2 else None,
        "se_mls": selinux[3] if lsel > 3 else None,
        "size": int(size),
        "name": path,
        "date": date,
    }
    if link:
        result["link"] = link
    return result


PASS_KEYS = set(["name", "total"])
DELAYED_KEYS = ["entries", "files", "dirs", "specials"]


class Directory(dict):
    def __init__(self, name, total, body):
        data = dict.fromkeys(DELAYED_KEYS)
        data["name"] = name
        data["total"] = total
        self.body = body
        self.loaded = False
        super(Directory, self).__init__(data)

    def iteritems(self):
        if not self.loaded:
            self._load()
        return six.iteritems(super(Directory, self))

    def items(self):
        if not self.loaded:
            self._load()
        return super(Directory, self).items()

    def values(self):
        if not self.loaded:
            self._load()
        return super(Directory, self).values()

    def get(self, key, default=None):
        if not self.loaded:
            self._load()
        return super(Directory, self).get(key, default)

    def _load(self):
        dirs = []
        ents = {}
        files = []
        specials = []
        for line in self.body:
            # we can't split(None, 5) here b/c rhel 6/7 selinux lines only have
            # 4 parts before the path, and the path itself could contain
            # spaces. Unfortunately, this means we have to split the line again
            # below
            parts = line.split(None, 4)
            perms = parts[0]
            typ = perms[0]
            entry = {
                "type": typ,
                "perms": perms[1:]
            }
            if parts[1][0].isdigit():
                # We have to split the line again to see if this is a RHEL8
                # selinux stanza. This assumes that the context section will
                # always have at least two pieces separated by ':'.
                if ":" in line.split()[4]:
                    rest = parse_rhel8_selinux(parts[1:])
                else:
                    rest = parse_non_selinux(parts[1:])
            else:
                rest = parse_selinux(parts[1:])

            # Update our entry and put it into the correct buckets
            # based on its type.
            entry.update(rest)
            entry["raw_entry"] = line
            entry["dir"] = self["name"]
            nm = entry["name"]
            ents[nm] = entry
            if typ not in "bcd":
                files.append(nm)
            elif typ == "d":
                dirs.append(nm)
            elif typ in "bc":
                specials.append(nm)

        self.update({"entries": ents,
                     "files": files,
                     "dirs": dirs,
                     "specials": specials})

        self.loaded = True
        del self.body

    def __getitem__(self, key):
        if self.loaded or key in PASS_KEYS:
            return super(Directory, self).__getitem__(key)
        self._load()
        return super(Directory, self).__getitem__(key)


def parse(lines, root=None):
    """
    Parses a list of lines from ls into dictionaries representing their
    components.

    Args:
        lines (list): A list of lines generated by ls.
        root (str): The directory name to be used for ls output stanzas that
            don't have a name.

    Returns:
        A dictionary representing the ls output. It's keyed by the path
        containing each ls stanza.
    """
    doc = {}
    entries = []
    name = None
    total = None
    for line in lines:
        line = line.strip()
        if not line:
            continue
        if line and line[0] == "/" and line[-1] == ":":
            if name is None:
                name = line[:-1]
                if entries:
                    d = Directory(name, total or len(entries), entries)
                    doc[root] = d
                    total = None
                    entries = []
            else:
                d = Directory(name, total or len(entries), entries)
                doc[name or root] = d
                total = None
                entries = []
                name = line[:-1]
            continue
        if line.startswith("total"):
            total = int(line.split(None, 1)[1])
            continue
        entries.append(line)
    name = name or root
    total = total if total is not None else len(entries)
    doc[name] = Directory(name, total, entries)
    return doc



'''
List files and dirs under ``/var/lib/nova/instances``
=====================================================

The parsers class in this module uses base parser class
``CommandParser`` & ``FileListing`` to list files & directories.

Parsers included in this modules are:

LsRVarLibNovaInstances - command ``ls -laR /var/lib/nova/instances``
---------------------------------------------------------------------

LsVarLibNovaInstances - command ``ls -laRZ /var/lib/nova/instances``
---------------------------------------------------------------------
'''
from insights import CommandParser, FileListing, parser
from insights.specs import Specs


@parser(Specs.ls_R_var_lib_nova_instances)
class LsRVarLibNovaInstances(CommandParser, FileListing):
    '''The class ``LsVarLibNovaInstances`` don't show file size when the
    flag `-Z` is used. This class parses the output of ``ls -laR
    /var/lib/nova/instances`` to output file listing with file size.

    Note: This issue is not seen in GNU coreutils-8.29. When the
    coreutils package is updated to 8.29 on RHEL7, this parser class
    can be deprecated.

    Typical output of the ``ls -laR /var/lib/nova/instances`` command is::

        /var/lib/nova/instances:
        total 4
        drwxr-xr-x. 5 nova nova  97 Feb 20  2017 .
        drwxr-xr-x. 9 nova nova 111 Feb 17  2017 ..
        drwxr-xr-x. 2 nova nova  54 Feb 17  2017 _base
        -rw-r--r--. 1 nova nova  44 May 26  2017 compute_nodes
        drwxr-xr-x. 2 nova nova  54 Feb 17  2017 e560e649-41fd-46a2-a3d2-5f4750ba2bb4
        drwxr-xr-x. 2 nova nova  93 Feb 17  2017 locks

        /var/lib/nova/instances/_base:
        total 18176
        drwxr-xr-x. 2 nova nova       54 Feb 17  2017 .
        drwxr-xr-x. 5 nova nova       97 Feb 20  2017 ..
        -rw-r--r--. 1 qemu qemu 41126400 May 26  2017 faf1184c098da91e90290a920b8fab1ee6e1d4c4

        /var/lib/nova/instances/e560e649-41fd-46a2-a3d2-5f4750ba2bb4:
        total 2104
        drwxr-xr-x. 2 nova nova      54 Feb 17  2017 .
        drwxr-xr-x. 5 nova nova      97 Feb 20  2017 ..
        -rw-r--r--. 1 qemu qemu   48957 Feb 20  2017 console.log
        -rw-r--r--. 1 qemu qemu 2097152 Feb 20  2017 disk
        -rw-r--r--. 1 nova nova      79 Feb 17  2017 disk.info

        /var/lib/nova/instances/locks:
        total 0
        drwxr-xr-x. 2 nova nova 93 Feb 17  2017 .
        drwxr-xr-x. 5 nova nova 97 Feb 20  2017 ..
        -rw-r--r--. 1 nova nova  0 Feb 17  2017 nova-faf1184c098da91e90290a920b8fab1ee6e1d4c4
        -rw-r--r--. 1 nova nova  0 Feb 17  2017 nova-storage-registry-lock


    Example:

        >>> ls_r_var_lib_nova_instances.dir_entry('/var/lib/nova/instances/e560e649-41fd-46a2-a3d2-5f4750ba2bb4', 'console.log')['size']
        48957
    '''
    pass


@parser(Specs.ls_var_lib_nova_instances)
class LsVarLibNovaInstances(CommandParser, FileListing):
    '''Parses the output of ``ls -laRZ /var/lib/nova/instances`` command
    which provides the SELinux directory listings of the '/var/lib/nova/instances' directory.

    The ``ls -laRZ /var/lib/nova/instances`` command provides
    information for the SELinux directory listing of the ``/var/lib/nova/instances`` directory.

    Typical output of the ``ls -laRZ /var/lib/nova/instances`` command is::

        /var/lib/nova/instances/:
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 11415c6c-a2a5-45f0-a198-724246b96631
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 _base
        -rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 compute_nodes
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 locks

        /var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631:
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
        -rw-------. root root system_u:object_r:nova_var_lib_t:s0 console.log
        -rw-r--r--. qemu qemu system_u:object_r:svirt_image_t:s0:c92,c808 disk
        -rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 disk.info

        /var/lib/nova/instances/_base:
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
        -rw-r--r--. qemu qemu system_u:object_r:virt_content_t:s0 572dfdb7e1d9304342cbe1fd5e3da4ff2e55c7a6

        /var/lib/nova/instances/locks:
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 .
        drwxr-xr-x. nova nova system_u:object_r:nova_var_lib_t:s0 ..
        -rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-572dfdb7e1d9304342cbe1fd5e3da4ff2e55c7a6
        -rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-storage-registry-lock

    Examples:

        >>> '/var/lib/nova/instances/' in ls_var_lib_nova_instances
        True
        >>> ls_var_lib_nova_instances.files_of('/var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631')
        ['console.log', 'disk', 'disk.info']
        >>> ls_var_lib_nova_instances.listings['/var/lib/nova/instances/11415c6c-a2a5-45f0-a198-724246b96631']['entries']['console.log']['se_type'] != 'nova_var_lib_t'
        False
        >>> len(ls_var_lib_nova_instances.listings['/var/lib/nova/instances/locks'])
        6
        >>> ls_var_lib_nova_instances.dir_entry('/var/lib/nova/instances/locks', 'nova-storage-registry-lock')['raw_entry']
        '-rw-r--r--. nova nova system_u:object_r:nova_var_lib_t:s0 nova-storage-registry-lock'
    '''
    pass



import pytest
from insights.util.file_permissions import FilePermissions
from insights.core import FileListing
from insights.tests import test_file_listing, context_wrap

PERMISSIONS_TEST_EXCEPTION_VECTORS = [
    ('-rw------ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('bash: ls: command not found', True),
    ('-rw------ 1 root root 762 Se', True),
    ('-rw------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------. 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------@ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------+ 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------* 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw-------asdf 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False),
    ('-rw------- 1 ro:t root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rw------- 1 root r:ot 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rwasdfas- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('-rwx/----- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True),
    ('/usr/bin/ls: cannot access /boot/grub2/grub.cfg: No such file or directory', True),
    ('cannot access /boot/grub2/grub.cfg: No such file or directory', True),
    ('No such file or directory', True),
    ('adsf', True),
]

PERMISSIONS_TEST_VECTORS = [
    ('-rw------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False,
     'rw-', '---', '---', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     True, True),
    ('-rw------- 1 root root 762 Sep 23 002 /a path/with/spaces everywhere', False,
     'rw-', '---', '---', 'root', 'root', '/a path/with/spaces everywhere',
     True, True,
     True, True),
    ('-rw------- 1 root root 762 Sep 23 002 no_slash_here', False,
     'rw-', '---', '---', 'root', 'root', 'no_slash_here',
     True, True,
     True, True),
    ('-rw-------. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config', False,
     'rw-', '---', '---', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     True, True),
    ('-rw-rw-rw-. 1 root root 4308 Apr 22 15:57 /etc/ssh/sshd_config', False,
     'rw-', 'rw-', 'rw-', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     False, False),
    ('-rw-rw---- 1 root user 762 Sep 23 002 /etc/ssh/sshd_config', True,
     'rw-', 'rw-', '---', 'root', 'user', '/etc/ssh/sshd_config',
     True, False,
     False, False),
    ('-rw------- 1 root user 762 Sep 23 002 /etc/ssh/sshd_config', False,
     'rw-', '---', '---', 'root', 'user', '/etc/ssh/sshd_config',
     True, False,
     True, True),
    ('-rw------- 1 user root 762 Sep 23 002 /etc/ssh/sshd_config', False,
     'rw-', '---', '---', 'user', 'root', '/etc/ssh/sshd_config',
     False, False,
     False, False),
    ('-rw-rw---- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True,
     'rw-', 'rw-', '---', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     True, True),
    ('-rw-rw-r-- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True,
     'rw-', 'rw-', 'r--', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     False, True),
    ('-rw-rw--w- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', True,
     'rw-', 'rw-', '-w-', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     True, False),
    ('---------- 1 root root 762 Sep 23 002 /etc/ssh/sshd_config', False,
     '---', '---', '---', 'root', 'root', '/etc/ssh/sshd_config',
     True, True,
     True, True),
]


def test_permissions():
    for vector in PERMISSIONS_TEST_VECTORS:
        (line, with_group,
         permissions_owner, permissions_group, permissions_other, owner, group, path,
         owned_by_root_user, owned_by_root_user_and_group,
         only_root_can_read, only_root_can_write) = vector
        p = FilePermissions(line)
        assert p.perms_owner == permissions_owner
        assert p.perms_group == permissions_group
        assert p.perms_other == permissions_other
        assert p.owner == owner
        assert p.group == group
        assert p.owned_by('root', also_check_group=False) == owned_by_root_user
        assert p.owned_by('root', also_check_group=True) == owned_by_root_user_and_group
        assert p.only_root_can_read(root_group_can_read=with_group) == only_root_can_read
        assert p.only_root_can_write(root_group_can_write=with_group) == only_root_can_write
        assert p.all_zero() == all((p.perms_owner == '---', p.perms_group == '---',
                                    p.perms_other == '---'))
        assert p.owner_can_read() == ('r' in p.perms_owner)
        assert p.owner_can_write() == ('w' in p.perms_owner)
        assert p.owner_can_only_read() == ('r--' == p.perms_owner)
        assert p.group_can_read() == ('r' in p.perms_group)
        assert p.group_can_write() == ('w' in p.perms_group)
        assert p.group_can_only_read() == ('r--' == p.perms_group)
        assert p.others_can_read() == ('r' in p.perms_other)
        assert p.others_can_write() == ('w' in p.perms_other)
        assert p.others_can_only_read() == ('r--' == p.perms_other)


def test_permissions_invalid():
    for vector in PERMISSIONS_TEST_EXCEPTION_VECTORS:
        garbage, should_raise = vector
        if should_raise:
            with pytest.raises(ValueError):
                FilePermissions(garbage)
        else:
            # shouldn't raise an exception
            FilePermissions(garbage)


def test_multiple_directories():
    dirs = FileListing(context_wrap(test_file_listing.MULTIPLE_DIRECTORIES))
    assert '/etc/sysconfig' in dirs
    assert 'cbq' in dirs.dirs_of('/etc/sysconfig')
    # drwxr-xr-x.  2 0 0   41 Jul  6 23:32 cbq
    obj = FilePermissions.from_dict(dirs.path_entry('/etc/sysconfig/cbq'))
    assert hasattr(obj, 'name')
    assert obj.name == 'cbq'
    assert obj.perms_owner == 'rwx'
    assert obj.perms_group == 'r-x'
    assert obj.perms_other == 'r-x'



import doctest
from insights.parsers import ls_tmp
from insights.parsers.ls_tmp import LsTmp
from insights.tests import context_wrap

LS_TMP = """
drwxrwxr-x.  2 whuser   whuser         216 Jul  9 07:09 aws_sos
-rw-r--r--.  1 rauser   rauser        1123 Jul 10 00:00 clean-old-archive.log
-rw-rw-r--.  1 whuser   whuser        9620 Jul  9 07:09 daily-extraction-warehouse-run.log
-rw-rw-r--.  1 whuser   whuser       11214 Jul  9 07:09 dask_master.log
-rw-rw-r--.  1 whuser   whuser       27091 Jul  9 07:09 dask_worker.log
-rw-r--r--.  1 user10   user10          29 Jul 10 00:00 date.out
-rw-r--r--.  1 rauser   rauser   325933528 Jul 10 00:18 delete-bad-pods.log
-rw-rw-r--.  1 whuser   whuser           0 Jul  9 07:08 extraction_driver.log
drwxrwxrwt.  2 root     root             6 Mar 27  2017 .font-unix
drwxrwxr-x.  3 user1    user1           17 Oct 28  2019 hadoop-user1
drwxr-xr-x.  2 user1    user1           32 Jul  4 18:29 hsperfdata_user1
drwxr-xr-x.  2 rauser   rauser           6 Jul 10 00:00 hsperfdata_rauser
drwxr-xr-x.  2 root     root             6 Jul  1 14:53 hsperfdata_root
drwxrwxrwt.  2 root     root             6 Mar 27  2017 .ICE-unix
srw-rw----.  1 user10   user10           0 Jan  6  2020 lh_pair
srw-rw----.  1 user10   user10           0 Jan 28 11:24 lh_pair_ex
"""

path = 'insights_commands/ls_-la_.tmp'


def test_ls_tmp():
    ls_tmp = LsTmp(context_wrap(LS_TMP, path=path))
    assert len(ls_tmp.listing_of("/tmp")) == 16
    assert len(ls_tmp.dirs_of("/tmp")) == 7
    assert len(ls_tmp.files_of("/tmp")) == 9
    expected = sorted(
        [
            'clean-old-archive.log', 'daily-extraction-warehouse-run.log',
            'dask_master.log', 'dask_worker.log', 'date.out',
            'delete-bad-pods.log', 'extraction_driver.log', 'lh_pair',
            'lh_pair_ex'
        ]
    )
    actual = ls_tmp.listings.get("/tmp")['files']
    assert actual == expected


def test_doc_examples():
    env = {'ls_tmp': LsTmp(context_wrap(LS_TMP, path=path))}
    failed, total = doctest.testmod(ls_tmp, globs=env)
    assert failed == 0



"""
RpmPkgs - datasource ``rpm_pkgs``
=================================

Parsers for CVE-2021-35937, CVE-2021-35938, and CVE-2021-35939.
For more information, see the ``rpm_pkgs`` datasource.
"""

from insights import Parser, parser
from insights.specs import Specs
from insights.util import deprecated


@parser(Specs.rpm_pkgs)
class RpmPkgs(Parser):
    """
    .. warning::
        This class is deprecated, please use
        :py:class:`insights.parsers.rpm_pkgs.RpmPkgsWritable` instead.

    Class for enabling the data from the ``rpm_pkgs`` datasource.

    Sample input of this parser is::

        ["httpd-core|httpd-core-2.4.53-7.el9.x86_64|Red Hat, Inc."]

    Sample output of this parser is::

        ["httpd-core"]

    Examples:

        >>> type(rpm_pkgs)
        <class 'insights.parsers.rpm_pkgs.RpmPkgs'>
        >>> rpm_pkgs.packages
        ['httpd-core']
    """
    def __init__(self, *args, **kwargs):
        deprecated(
            RpmPkgs,
            "Please use the `RpmPkgsWritable` instead.",
            "3.2.25"
        )
        super(RpmPkgs, self).__init__(*args, **kwargs)

    def parse_content(self, content):
        self.packages = [pkg.split('|')[0] for pkg in content]


@parser(Specs.rpm_pkgs)
class RpmPkgsWritable(Parser):
    """
    Class for enabling the data from the ``rpm_pkgs`` datasource.
    It replaces the original RpmPkgs parser.

    Sample input of this parser is::

        ["httpd-core|httpd-core-2.4.53-7.el9.x86_64|Red Hat, Inc."]

    Sample output of this parser is::

        [("httpd-core", "httpd-core-2.4.53-7.el9.x86_64", "Red Hat, Inc.")]

    Examples:

        >>> type(rpm_pkgs_writable)
        <class 'insights.parsers.rpm_pkgs.RpmPkgsWritable'>
        >>> rpm_pkgs_writable.packages
        [('httpd-core', 'httpd-core-2.4.53-7.el9.x86_64', 'Red Hat, Inc.')]
    """

    def parse_content(self, content):
        self.packages = [tuple(pkg.split('|')) for pkg in content]



"""
Crash log ``vmcore-dmesg.txt``
==============================

This module provides ``vmcore-dmesg.txt`` crash logs that are stored in ``/var/crash/[host]-YYYY-MM-DD-HH:MM:SS/``.

Exemplary ``vmcore-dmesg.txt`` looks like::

    [  345.691798] device-mapper: raid: Failed to read superblock of device at position 0
    [  345.693497] device-mapper: raid: Discovered old metadata format; upgrading to extended metadata format
    [  345.701166] md/raid1:mdX: active with 1 out of 2 mirrors
    [  345.726870] BUG: unable to handle kernel NULL pointer dereference at 00000000000005ec
    [  345.727782] IP: [<ffffffffc0852ffb>] read_balance+0x1db/0x4e0 [raid1]
    [  345.728570] PGD 0
    [  345.729279] Oops: 0000 [#1] SMP
    [  345.729950] Modules linked in: raid1 dm_raid raid456 async_raid6_recov async_memcpy async_pq raid6_pq async_xor xor async_tx
    [  345.734752]  drm_kms_helper syscopyarea sysfillrect sysimgblt fb_sys_fops ttm drm bfa ahci crct10dif_pclmul crct10dif_common
    [  345.737334] CPU: 6 PID: 952 Comm: systemd-udevd Kdump: loaded Not tainted 3.10.0-862.9.1.el7.x86_64 #1
    [  345.738198] Hardware name: Dell Inc. PowerEdge R730xd/072T6D, BIOS 2.4.3 01/17/2017
    [  345.739334] task: ffff9f888d030fd0 ti: ffff9f86b76b4000 task.ti: ffff9f86b76b4000

The class inherites LogFileOutput and doesn't implement new methods/attributes.
vmcore_dmesg defined in Specs allowes multioutput parsing. Filtering is allowed.

Examples:
    >>> vmcore_dmesg.get("Modules linked in:")
    [{'raw_message': '[  345.729950] Modules linked in: raid1 dm_raid raid456 async_raid6_recov async_memcpy async_pq raid6_pq async_xor xor async_tx'}]
    >>> vmcore_dmesg.get("device-mapper")
    [{'raw_message': '[  345.691798] device-mapper: raid: Failed to read superblock of device at position 0'}, {'raw_message': '[  345.693497] device-mapper: raid: Discovered old metadata format; upgrading to extended metadata format'}]
"""

from .. import parser, LogFileOutput
from insights.specs import Specs


@parser(Specs.vmcore_dmesg)
class VMCoreDmesg(LogFileOutput):
    """
    This class parses data in ``/var/crash/[host]-YYYY-MM-DD-HH:MM:SS/vmcore-dmesg.txt``.
    Inherited from LogFileOutput. See LogFileOutput for details. No new methods or attributes.
    MultiOutput is allowed. Filtering is allowed.
    """
    pass



import doctest
import pytest

from insights.core.exceptions import ParseException, SkipComponent
from insights.parsers import subscription_manager
from insights.parsers.subscription_manager import SubscriptionManagerFacts
from insights.tests import context_wrap

INPUT_NORMAL_1 = """
aws_instance_id: 567890567890
network.ipv6_address: ::1
uname.sysname: Linux
uname.version: #1 SMP PREEMPT Fri Sep 2 16:07:40 EDT 2022
virt.host_type: rhev, kvm
virt.is_guest: True
""".strip()

INPUT_NG_1 = """
XYC
Release not set
""".strip()

INPUT_NG_2 = ""


def test_subscription_manager_facts():
    ret = SubscriptionManagerFacts(context_wrap(INPUT_NORMAL_1))
    for line in INPUT_NORMAL_1.splitlines():
        key, value = line.split(': ', 1)
        assert ret[key] == value


def test_subscription_manager_release_show_ng():
    with pytest.raises(ParseException):
        SubscriptionManagerFacts(context_wrap(INPUT_NG_1))

    with pytest.raises(SkipComponent):
        SubscriptionManagerFacts(context_wrap(INPUT_NG_2))


def test_doc_examples():
    env = {
            'rhsm_facts':
            SubscriptionManagerFacts(context_wrap(INPUT_NORMAL_1)),
          }
    failed, total = doctest.testmod(subscription_manager, globs=env)
    assert failed == 0



import doctest
from insights.parsers import pcp_openmetrics_log
from insights.tests import context_wrap


PCP_OPENMETRICS_LOG = """
[Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store 55% {:
[Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store }:
[Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store 95% {:
[Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store }:
[Mon Jul 11 07:59:32] pmdaopenmetrics(1733) Error: cannot parse/store };:
""".strip()


def test_pcp_openmetrics_log():
    log = pcp_openmetrics_log.PcpOpenmetricsLog(context_wrap(PCP_OPENMETRICS_LOG))
    assert "Error: cannot parse/store" in log
    assert len(log.get('pmdaopenmetrics')) == 5


def test_documentation():
    failed_count, tests = doctest.testmod(
        pcp_openmetrics_log,
        globs={'log': pcp_openmetrics_log.PcpOpenmetricsLog(context_wrap(PCP_OPENMETRICS_LOG))}
    )
    assert failed_count == 0



"""
GlanceApiLog - file ``/var/log/glance/api.log``
===============================================

Module for parsing the log files for Glance

"""

from .. import LogFileOutput, parser
from insights.specs import Specs


@parser(Specs.glance_api_log)
class GlanceApiLog(LogFileOutput):
    """Class for parsing ``/var/log/glance/api.log`` file.

    Typical content of ``api.log`` file is::

        2016-11-09 14:50:44.281 26656 INFO glance.common.wsgi [-] Started child 14826
        2016-11-09 14:50:44.445 14826 INFO eventlet.wsgi.server [-] (14826) wsgi starting up on http://172.18.0.13:9292
        2016-11-09 14:50:44.454 14826 INFO eventlet.wsgi.server [-] (14826) wsgi exited, is_accepting=True
        2016-11-09 14:50:44.470 14826 INFO glance.common.wsgi [-] Child 14826 exiting normally
        2016-11-09 14:50:49.032 14863 WARNING oslo_config.cfg [-] Option "rpc_backend" from group "DEFAULT" is deprecated for removal.  Its value may be silently ignored in the future.
        2016-11-09 14:50:49.529 14863 INFO glance.common.wsgi [-] Starting 4 workers
        2016-11-09 14:50:49.539 14863 INFO glance.common.wsgi [-] Started child 15049
        2016-11-09 14:50:49.550 14863 INFO glance.common.wsgi [-] Started child 15054
        2016-11-09 14:50:49.552 15054 INFO eventlet.wsgi.server [-] (15054) wsgi starting up on http://172.18.0.13:9292
        2016-11-09 14:50:49.561 15049 INFO eventlet.wsgi.server [-] (15049) wsgi starting up on http://172.18.0.13:9292
        2016-11-09 14:50:49.719 14863 INFO glance.common.wsgi [-] Started child 15097
        2016-11-09 14:50:49.726 14863 INFO glance.common.wsgi [-] Started child 15101
        2016-11-09 14:50:49.727 15101 INFO eventlet.wsgi.server [-] (15101) wsgi starting up on http://172.18.0.13:9292
        2016-11-09 14:50:49.730 15097 INFO eventlet.wsgi.server [-] (15097) wsgi starting up on http://172.18.0.13:9292

    .. note::
        Please refer to its super-class :class:`insights.core.LogFileOutput`
    """
    pass
